/***************************************************************************
 *   Copyright (C) 2005 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <iostream>
#include <iomanip>
using namespace std;

#include <math.h>

#include "Analyzer.h"

#ifndef QSETTINGS
#include "QSettings.h"
#else
#include QSETTINGS
#endif

// helper called by the constructors
void Analyzer::constructor(int nlevel, int ncoefmin, Wavelet* w, double* wght, bool discardOlderData, bool averageAllFrames, bool monitorCorrelation, bool useMemory) {

    wavelet = w->ref();
    this->discardOlderData = discardOlderData;
    this->averageAllFrames = averageAllFrames;
    this->monitorCorrelation = monitorCorrelation;
    this->useMemory = useMemory;

    // make local copy of weights
    weights = new double[nlevel];
    memcpy(weights, wght, nlevel*sizeof(double));

    nframes = 1 << nlevel;
    nframesInv = 1.0 / (double)nframes;
    filter_index_mask = nframes - 1;

    delay = (nframes - 1) * (w->support - 2);
    newMFAFrameFirstIdx = data.capacity() - delay - nframes;

    nextLevel = new DecompositionLevel(nlevel-1, ncoefmin, w);
    nextLevelWaiting = new DecompositionLevel(nlevel-1, ncoefmin, w);

    this->nlevel = nlevel;

    newDataDWTIdx = data.capacity() - wavelet->support;

    // Start with the initialization version
    // Then switch to full-buffer regime when this happens
    newDataDWTPtr = &Analyzer::newDataDWTInitPhase;

    // Initialize the filters now we have the number of levels and wavelet
    // This may create a whole new filter set, or re-use an already existing one
    scaleDeltaFilterArray = ScaleDeltaFilterArray::getInstance(w, nlevel);
    // Cache the internal array for direct access efficiency
    scaleDeltaFilter = scaleDeltaFilterArray->getArray();

    // hq re-initialized each use, except when averaging over all frames
    hq = new double[NUM_Q];
    memset(hq, 0, NUM_Q * sizeof(double));
    if (monitorCorrelation) {
        squaredCorrelation = new double[NUM_Q];
        memset(squaredCorrelation, 0, NUM_Q * sizeof(double));
    } else squaredCorrelation =0;

    // Compute constants for H(q) computations from power sums

    // Now use weighted least squares fitting formula
    double m1x = 0.0;
    double m2x = 0.0;
    double sw = 0.0;
    for (int s=1; s<=nlevel; ++s) {
        double x = s;// * log(2.0);
        m1x += x * weights[s-1];
        m2x += x*x * weights[s-1];
        sw += weights[s-1];
    }
    double commonConst = sw * m2x - m1x * m1x;

    // variables used for correlation coefficient computation
    correlationCoefficientDenominatorTerm1 = m2x - m1x * m1x / sw;
    meanX = m1x / sw;
    sumOfWeightsInv = 1.0 / sw;

    // Include all frames averaging into constant when this is the case
    // => save one average computation!
    // cannot be used when monitoring Correlations, we need to average the Correlations themselves
    if (averageAllFrames && (!monitorCorrelation)) commonConst *= nframes;

    // Compute constant
    powerSumToHqConstant = new double*[nlevel+1];
    powerSumToHqConstant[0] = 0;
    
    for (int s=1; s<=nlevel; ++s) {
        powerSumToHqConstant[s] = new double[NUM_Q]; // number of q
        double x = s;// * log(2.0);
        
        for (int i=0; i<NUM_Q; ++i) {
            double q = GET_Q_FOR_INDEX(i);
            powerSumToHqConstant[s][i] = weights[s-1] * (sw * x - m1x) / (q * commonConst);
        }
    }
}

Analyzer::~Analyzer() {
    // cleanup hq
    delete [] hq;

    // cleanup constants
    for (int s=1; s<=nlevel; ++s) {
        delete [] powerSumToHqConstant[s];
    }
    delete [] powerSumToHqConstant;

    delete [] weights;

    // cleanup filters
    scaleDeltaFilterArray->unref();

    // cleanup levels (recursive)
    delete nextLevel;
    delete nextLevelWaiting;

    if (monitorCorrelation) delete [] squaredCorrelation;

    // cleanup history (if needed)
    for (int i=0; i<hqHistory.size(); ++i) delete [] hqHistory[i];

    // cleanup Correlation history (if needed)
    for (int i=0; i<correlationHistory.size(); ++i) delete [] correlationHistory[i];

    // release shared instance
    wavelet->unref();
}

void Analyzer::reset() {

    // clear histories
    for (int i=0; i<hqHistory.size(); ++i) delete [] hqHistory[i];
    for (int i=0; i<correlationHistory.size(); ++i) delete [] correlationHistory[i];
    hqHistory.size() = 0;
    correlationHistory.size() = 0;

    // hq re-initialized too
    memset(hq, 0, NUM_Q * sizeof(double));
    if (squaredCorrelation) memset(squaredCorrelation, 0, NUM_Q * sizeof(double));

    // Clear data
    data.size() = 0;

    // Start with the initialization version
    // Then switch to full-buffer regime when this happens
    newDataDWTPtr = &Analyzer::newDataDWTInitPhase;

    // recursive reset levels
    if (nextLevel) nextLevel->reset();
    if (nextLevelWaiting) nextLevelWaiting->reset();
}

Analyzer::Analyzer(int nlevel, int ncoefmin, Wavelet* w, double* weights, bool discardOlderData, bool averageAllFrames, bool monitorCorrelation, bool useMemory) : data((1 << nlevel) * ncoefmin + (w->support - 2) * ((1 << nlevel) - 1)), hqHistory(1 << nlevel), correlationHistory(1 << nlevel) {

    // Use user-specified weights for exponential fit
    constructor(nlevel, ncoefmin, w, weights, discardOlderData, averageAllFrames, monitorCorrelation, useMemory);
}

Analyzer::Analyzer(int nlevel, int ncoefmin, Wavelet* w, bool discardOlderData, bool averageAllFrames, bool monitorCorrelation, bool useMemory) : data((1 << nlevel) * ncoefmin + (w->support - 2) * ((1 << nlevel) - 1)), hqHistory(1 << nlevel), correlationHistory(1 << nlevel) {

    // Use weighting in least squares estimate for exponential fit y = A * exp(B*x)   (eq.1)
    // or equivalently log y = B*x + log A     (eq.2)
    // Here we want to estimate B = h(q), with x = s * log(2). We don't care for A, so using the powerSums instead
    // of the full F(s) = (1/N * power_sum) ^ q is possible. Eq. 2 shows we also need to divide by q to get
    // a correct B back  (and similarly adding log N would be needed for A, but we don't care for A).
    // Simple least squares (no weighting) minimizes sum_over_all_(x,y)[ (log y - (B*x + log A))^2 ]
    // Unfortunately this gives more influence to lower scales due to the log transform
    // A possible way is to give all (x,y) a roughly identical influence by using the y as weights (as they
    // are close to the final exponential) : min sum_over_all_(x,y)[ y * (log y - (B*x + log A))^2 ]
    // Unfortunately this is still not quite correct, since y is only an estimate for the fitted exponential
    // and there may be large variations giving too much weight to a y (hence to that variation...).
    // Nevertheless, this is still much better than not using any weight at all. Other extentions would perhaps
    // consider using the estimated value of B to get a better weighting scheme than y, and iterate till
    // convergence.
    // In our case, using the y as weights would prevent any pre-computation. This is unfortunate as we don't
    // wan't to sacrifice too much speed for a slight increase in precision.
    // However, an exponential weigthing scheme does not cost more than unweighted scheme thanks to precomputation
    // => we might as well use it
    // The compromise chosen is to use the weights corresponding to a theoretical white noise fractal brownian motion "monofractal" with H=0.5.
    // Thus, we get an exponential weigthing scheme that is pre-computable. Numerical experiments shown that using the
    // real y compared to this weighting scheme wouldn't bring enough additional precision compared to the unweighted
    // scheme. Of course, depending on a specific application, you may want to sacrifice speed for that little extra, but
    // the author choosed not to make that computationally more expensive decision the default choice.
    double weights[nlevel];
    for (int s=1; s<=nlevel; ++s) {
        // Read above for explanation: exp(0.5 * x) with x = s * log(2)
        weights[s-1] = sqrt((double)(1 << s));
    }

    constructor(nlevel, ncoefmin, w, weights, discardOlderData, averageAllFrames, monitorCorrelation, useMemory);

}

/// Discrete Wavelet Transform the data into the next level
/// low and high band coefficients
void Analyzer::transform(CyclicArray& data, DecompositionLevel* level) {
    // Boundary coefficients won't be reconstructed,
    // there is no way around that if we don't wan't to introduce
    // boundary effects (which would have too much influence on
    // the scale/frequency analysis)
    const int max = data.size() - wavelet->support;
    double lphp[2];
    for (int i=0; i<=max; i+=2) {
        wavelet->convolve(data, i, lphp);
        level->lowPass[i>>1] = lphp[0];
        level->highPass[i>>1] = lphp[1];
    }
}

/// restore data from next level by inverse DWT
void Analyzer::restore(DecompositionLevel* level, CyclicArray& data) {
    const int max = level->lowPass.size()-(wavelet->support-2)/2;
    // Boundary coefficients are not reconstructed
    int idx = 0;
    for (; idx < wavelet->support-2; ++idx) data[idx] = 0;
    double evenodd[2];
    for (int i=0; i<max; ++i) {
        wavelet->reconstruct(level->lowPass, level->highPass, i, evenodd);
        data[idx++] = evenodd[0];
        data[idx++] = evenodd[1];
    }
    // other half of the discarded coefficients
    while (idx < data.size()) data[idx++] = 0;
}


double Analyzer::computeScaleDelta(int dataIndex, int scale, CyclicArray& hp) {
    const ScaleDeltaFilter& filter = scaleDeltaFilter[dataIndex & filter_index_mask];
    const int& ncoefs = filter.size[scale];
    double ret = 0.0;
    int hpidx = (dataIndex >> scale) - ncoefs + 1;
    double* hplin = hp.linear(hpidx,ncoefs);
    if (hplin) {
        for (int i = 0; i < ncoefs; ++i) {
            ret += filter.coefs[scale][i] * hplin[i];
        }
        return ret;
    }
    for (int i = 0; i < ncoefs; ++i) {
        ret += filter.coefs[scale][i] * hp[hpidx+i];
    }
    return ret;
}

DecompositionLevel* Analyzer::newDataDWTInitPhase(double newData) {
    if (data.isFull()) {
        // remove old data
        data.pop_front();
        // Switch to full buffer version for next time
        newDataDWTPtr = &Analyzer::newDataDWTFullRegime;
    }
    // add new data
    data.push_back(newData);
    // Swap buffers
    DecompositionLevel* tmplp = nextLevelWaiting;
    nextLevelWaiting = nextLevel;
    nextLevel = tmplp;
    // update DWT coefficients for feeding next level recursively
    const int idx = data.size() - wavelet->support;
    if (idx<0) return 0;
    double lphp[2];
    wavelet->convolve(data, idx, lphp);
    return nextLevel->newDataUpdate(lphp);
}

DecompositionLevel* Analyzer::newDataDWTFullRegime(double newData) {
    // add new data
    data.push_cyclic(newData);
    // Swap buffers
    DecompositionLevel* tmplp = nextLevelWaiting;
    nextLevelWaiting = nextLevel;
    nextLevel = tmplp;
    // update DWT coefficients for feeding next level recursively
    double lphp[2];
    wavelet->convolve(data, newDataDWTIdx, lphp);
    return nextLevel->newDataUpdate(lphp);
}



int Analyzer::newDataMFA(double data) {

    double delta[nframes];
    DecompositionLevel* level;

    // PART 1: Remove old point contribution to the MFA spectrum (old frame)
    //         This is done except if user asked to compute over the whole
    //         series, from the beginning. This whole series mode may be
    //         appropriate for stationary series as it averages over more data.
    //         However this is inappropriate for non-stationary series. Computation
    //         over only the most recent data (and removal of old points) is thus the default.
    //--------

    if (discardOlderData) {
        level = nextLevel;

        // Assume 0-bitfield encodes for double 0.0
        memset(delta, 0, nframes * sizeof(double));
        // Proceed level by level & only update what's necessary
        for (int scale = 1; scale<=nlevel; ++scale) {

            // Some key indices for data re-use & computation
            const int split_idx_1 = (1<<scale);

            // Stop if a level is not ready. By construction, upper levels aren't either
            // A level is ready when it's power-sum was set. Otherwise, there is nothing to remove.
            if (level->powerSum==0) break;

            // Step 1: Update the delta. Use the fact that points to remove are at index delay
            //         from the array start (symmetric transform)
            CyclicArray& hp = level->highPass;

            // CPU version computes ALL the nframes delta, so as to pass them on to higher level.
            // This is also the case for highest level in MEM version, as no data can be reused then
            if ((!useMemory) || (scale==nlevel)) {
                for (int i=0; i<nframes; ++i) delta[i] += computeScaleDelta(delay+i, scale, hp);

            // MEM version computes only what could not be re-used from last call, but storing
            // all these delta quickly uses a lot of memory for large number of levels
            } else {
                // Thanks to previous test, split_idx_2 >= split_idx_1
                const int split_idx_2 = nframes - split_idx_1;
                // re-use existing values
                for (int i=0; i<split_idx_2; ++i) {
                    delta[i] += level->cachedOldDataDelta[i];
                }
                // update existing values for next time
                for (int i=split_idx_1; i<split_idx_2; ++i) {
                    level->cachedOldDataDelta[i-split_idx_1] = level->cachedOldDataDelta[i];
                }
                // compute missing ones
                for (int i=split_idx_2; i<nframes; ++i) {
                    double theDelta = computeScaleDelta(delay+i, scale, hp);
                    level->cachedOldDataDelta[i-split_idx_1] = theDelta;
                    delta[i] += theDelta;
                }
            }
    
            // Step 2: update the q-power series for each q: only the first terms should be
            //         updated, this particular level is called once every 2^s
            for (int i = 0; i<split_idx_1; ++i) {
                double term = fabs(delta[i]);
                // Process positive values
                minusExponentiationPos(term, level->powerSum);
#ifdef HAS_NEGATIVE_Q
                // Process negative values
                minusExponentiationNeg(1.0 / term, level->powerSum);
#endif
            }
    
            // really bad conditionning: tell user and provoke full recomputation
            for (int i=0; i<NUM_Q; ++i) {
                if (level->powerSum[i]<0) {
                    //cout << "Too many numerical errors. Please process full recomputations from time to time." << endl;
                    level->powerSum[i]=0;//throw "error";
                }
            }
            //cout << endl;
    
            level = level->nextLevel;
        }
    }
    
    // PART 2: Update the DWT and the current frame with the new data
    //--------
    DecompositionLevel* lastLevel = newDataDWT(data);

    // take care of initialization phase. Skip computations for now
    if (!lastLevel) return 2;

    // Check whether last level is ready or not
    // It is ready when it is full, and care has to be taken for the first
    // time it is so. Note: this isn't the optimal way to handle things,
    // in the sense that reconstruction at scale s<L could be done earlier
    // in the process for each point in the sum. But then, if all we care for
    // is the full spectrum (not partial ones), then all these computations
    // are useless
    if (lastLevel->powerSum==0) {
        // Compute the whole power sum the first time it can be done for
        // this frame. Use non-incremental version for this
        if (lastLevel->lowPass.isFull()) nonIncrementalMFA();
        // Otherwise, lastLevel->powerSum remains 0
        return (lastLevel->powerSum!=0) ? 1 : 2; // when true, h(q) were updated by non-incremental MFA
    }


    // PART 3: Add the new point contribution to the MFA spectrum (new frame)
    //--------

    // Assume 0-bitfield encodes for double 0.0
    memset(delta, 0, nframes * sizeof(double));

    // when not averaging over all frames, default is to use class HQ
    double * workHQ = hq;
    double * workSquaredCorrelation = squaredCorrelation;
    // Otherwise, maintain class HQ to averaged value
    if (averageAllFrames) {
        if (hqHistory.isFull()) {
            workHQ = hqHistory[0];
            for (int i=0; i<NUM_Q; ++i) hq[i] -= workHQ[i];
            hqHistory.push_cyclic(workHQ); //pushes itself at the same place!
        } else {
            hqHistory.push_back(workHQ = new double[NUM_Q]);
        }
        // process Correlation history the same way
        if (monitorCorrelation) {
            if (correlationHistory.isFull()) {
                workSquaredCorrelation = correlationHistory[0];
                for (int i=0; i<NUM_Q; ++i) squaredCorrelation[i] -= workSquaredCorrelation[i];
                correlationHistory.push_cyclic(workSquaredCorrelation); //pushes itself at the same place!
            } else {
                correlationHistory.push_back(workSquaredCorrelation = new double[NUM_Q]);
            }
        }
    }
    // initialize the h(q) computation
    memset(workHQ, 0, NUM_Q * sizeof(double));

    // initialize the Correlations if needed
    // Temporary use the array for the b values
    if (monitorCorrelation) memset(workSquaredCorrelation, 0, NUM_Q * sizeof(double));

    // Temporary place holders on the stack for each q
    double sumWY[NUM_Q] = EMPTY_Q_INITIALIZER;
    double sumWXY[NUM_Q] = EMPTY_Q_INITIALIZER;
    double sumWYY[NUM_Q] = EMPTY_Q_INITIALIZER;

    // Add new points to the new frame
    // Proceed level by level & only update what's necessary
    level = nextLevel;
    for (int scale = 1; scale<=nlevel; ++scale) {

        // Some key indices for data re-use & computation
        const int split_idx_1 = (1<<scale);
        const int split_idx_2 = nframes - split_idx_1;

        // Step 1: Update the delta.
        CyclicArray& hp = level->highPass;

        // CPU version computes ALL the nframes delta, so as to pass them on to higher level.
        // This is also the case for highest level in MEM version, as no data can be reused then
        if ((!useMemory) || (scale==nlevel)) {
            for (int i=0; i<nframes; ++i) delta[i] += computeScaleDelta(newMFAFrameFirstIdx+i, scale, hp);

        // MEM version computes only what could not be re-used from last call, but storing
        // all these delta quickly uses a lot of memory for large number of levels
        } else {
            // Thanks to previous test, split_idx_2 >= split_idx_1
            // re-use existing values
            for (int i=0; i<split_idx_2; ++i) {
                delta[i] += level->cachedNewDataDelta[i];
            }
            // update existing values for next time
            for (int i=split_idx_1; i<split_idx_2; ++i) {
                level->cachedNewDataDelta[i-split_idx_1] = level->cachedNewDataDelta[i];
            }
            // compute missing ones
            for (int i=split_idx_2; i<nframes; ++i) {
                double theDelta = computeScaleDelta(newMFAFrameFirstIdx+i, scale, hp);
                level->cachedNewDataDelta[i-split_idx_1] = theDelta;
                delta[i] += theDelta;
            }
        }

        // Step 2: update the q-power series for each q: only the last terms should be
        //         updated, this particular level is called once every 2^s
        for (int i = split_idx_2; i<nframes; ++i) {
            double term = fabs(delta[i]);
            // Process positive values
            plusExponentiationPos(term, level->powerSum);
#ifdef HAS_NEGATIVE_Q
            // Process negative values
            plusExponentiationNeg(1.0 / term, level->powerSum);
#endif
        }

        // when monitoring Correlations, we have to keep the logs for later use
        if (monitorCorrelation) {
            double weight = weights[scale-1];
            for (int i=0; i<NUM_Q; ++i) {
                double log2y;
#ifdef NO_YLOG2X_ASM
                log2y = log2(level->powerSum[i]);
#else
                asm(
                "fld1\n"
                "fxch\n"
                "fyl2x"    // powerSumToHqConstant[scale][i] * log2(level->powerSum[i]) in one instruction
                : "=t" (log2y)
                : "0" (level->powerSum[i])
                : "st(1)"
                );
#endif
                workHQ[i] += powerSumToHqConstant[scale][i] * log2y;
                double log2yW = log2y * weight;
                sumWY[i] += log2yW;
                sumWYY[i] += log2yW * log2y;
                sumWXY[i] += log2yW * scale;
            }
            level = level->nextLevel;
            continue;
        }

        // Step 3: update the h(q) with the power sums
        for (int i=0; i<NUM_Q; ++i) {
#ifdef NO_YLOG2X_ASM
            workHQ[i] += powerSumToHqConstant[scale][i] * log2(level->powerSum[i]);
#else
            asm(
            "fyl2x\n\t"    // powerSumToHqConstant[scale][i] * log2(level->powerSum[i]) in one instruction
            "faddl %[hq]" // loads and add workHQ[i]
            : "=t" (workHQ[i])
            : "u" (powerSumToHqConstant[scale][i]), "0" (level->powerSum[i]), [hq] "m" (workHQ[i])
            : "st(1)"
            );
#endif
        }

        level = level->nextLevel;
    }

    // Error-checking code
    for (int i=0; i<NUM_Q; ++i) {
        // Use 0.0 so it can be added & subtracted later on, NaN cannot
        if (!isnormal(workHQ[i])) workHQ[i] = 0.0;
    }

    // Complete the Correlation computation only if necessary
    if (!monitorCorrelation) {
        
        // When averaging over all frames, update current sum
        if (averageAllFrames) for (int i=0; i<NUM_Q; ++i) hq[i] += workHQ[i];
        
        // h(q) computation done! the new MFA spectrum is ready :)
        return 0;
    }


    // for each q
    for (int i=0; i<NUM_Q; ++i) {
        // Complete the correlation coefficient computation
        double cross_product = sumWXY[i] - meanX * sumWY[i];
        workSquaredCorrelation[i] = cross_product * cross_product / (correlationCoefficientDenominatorTerm1 * (sumWYY[i] - sumWY[i] * sumWY[i] * sumOfWeightsInv));
        // Use 0.0 so it can be added & subtracted later on, NaN cannot
        if (!isnormal(workSquaredCorrelation[i])) workSquaredCorrelation[i] = 0.0;
    }

    // compute averages over frames if necessary
    if (averageAllFrames) {
        for (int i=0; i<NUM_Q; ++i) {
            workSquaredCorrelation[i] *= nframesInv;
            workHQ[i] *= nframesInv;
            squaredCorrelation[i] += workSquaredCorrelation[i];
            hq[i] += workHQ[i];
        }
    }

    return 0;

}

void Analyzer::nonIncrementalMFA( )
{
    // Number of valid data points that can be reconstructed
    int npoints = data.capacity() - delay*2;

    // Assume 0-bitfield encodes for double 0.0
    double delta[npoints];
    memset(delta, 0, npoints * sizeof(double));

    // when not averaging over all frames, default is to use class HQ
    double * workHQ = hq;
    double * workSquaredCorrelation = squaredCorrelation;
    // Otherwise, maintain class HQ to averaged value
    if (averageAllFrames) {
        if (hqHistory.isFull()) {
            workHQ = hqHistory[0];
            for (int i=0; i<NUM_Q; ++i) hq[i] -= workHQ[i];
            hqHistory.push_cyclic(workHQ); //pushes itself at the same place!
        } else {
            hqHistory.push_back(workHQ = new double[NUM_Q]);
        }
        // process Correlation history the same way
        if (monitorCorrelation) {
            if (correlationHistory.isFull()) {
                workSquaredCorrelation = correlationHistory[0];
                for (int i=0; i<NUM_Q; ++i) squaredCorrelation[i] -= workSquaredCorrelation[i];
                correlationHistory.push_cyclic(workSquaredCorrelation); //pushes itself at the same place!
            } else {
                correlationHistory.push_back(workSquaredCorrelation = new double[NUM_Q]);
            }
        }
    }
    // initialize the h(q) computation
    memset(workHQ, 0, NUM_Q * sizeof(double));

    // initialize the Correlations if needed
    // Temporary use the array for the b values
    if (monitorCorrelation) memset(workSquaredCorrelation, 0, NUM_Q * sizeof(double));

    // Temporary place holders on the stack for each q
    double sumWY[NUM_Q] = EMPTY_Q_INITIALIZER;
    double sumWXY[NUM_Q] = EMPTY_Q_INITIALIZER;
    double sumWYY[NUM_Q] = EMPTY_Q_INITIALIZER;

    // Compute the whole power sums for this frame
    // First level needed before other levels for h(q) computations
    // Proceed level by level
    DecompositionLevel* level = nextLevel;
    for (int scale = 1; scale<=nlevel; ++scale) {

        // Step 1: Update the delta over the whole data range
        // The non-incremental version updates over
        // the whole range, no previous computation is re-used.
        // However, we have to prepare the data caches for later incremental updates

        CyclicArray& hp = level->highPass;

        // CPU version & highest level MEM version do not store any data
        if ((!useMemory) || (scale==nlevel)) {
            for (int i=0; i<npoints; ++i) delta[i] += computeScaleDelta(delay+i, scale, hp);

        // MEM version: cache results for future incremental updates
        } else {
            // Warning: these split variables correspond to the whole caches, unlike incremental version
            const int split_idx_1 = nframes - (1<<scale);
            const int split_idx_2 = npoints - split_idx_1;

// Define this symbol if you wish perfect match between the memory and CPU versions.
// Unfortunately the expense of using long double may on some systems outweight the
// speed advantage of MEM vs CPU. According to your system, it might then be more
// advantageous to stick to the CPU version!
#ifndef MEM_CPU_VERSIONS_PERFECT_MATCH
            if (!level->cachedOldDataDelta) level->cachedOldDataDelta = new double[split_idx_1];
            if (!level->cachedNewDataDelta) level->cachedNewDataDelta = new double[split_idx_1];
#else
            if (!level->cachedOldDataDelta) level->cachedOldDataDelta = new long double[split_idx_1];
            if (!level->cachedNewDataDelta) level->cachedNewDataDelta = new long double[split_idx_1];
#endif

            // Process cache for old points
            for (int i=0; i<split_idx_1; ++i) {
                double theDelta = computeScaleDelta(delay+i, scale, hp);
                level->cachedOldDataDelta[i] = theDelta;
                delta[i] += theDelta;
            }

            // No cache for middle data points
            for (int i=split_idx_1; i<split_idx_2; ++i) {
                delta[i] += computeScaleDelta(delay+i, scale, hp);
            }

            // Process cache for new points
            for (int i=split_idx_2; i<npoints; ++i) {
                double theDelta = computeScaleDelta(delay+i, scale, hp);
                level->cachedNewDataDelta[i-split_idx_2] = theDelta;
                delta[i] += theDelta;
            }
        }

        // Step 2: re-initialize the power sums
        // this may be used for initialization, or simply to recover
        // from potential numerical roundup errors after many updates
        if (level->powerSum==0) level->powerSum = new double[NUM_Q];
        memset(level->powerSum, 0, NUM_Q * sizeof(double));

        // Step 3: update the q-power series for each q, each delta of each data
        for (int i = 0; i<npoints; ++i) {
            double term = fabs(delta[i]);
            // Process positive values
            plusExponentiationPos(term, level->powerSum);
#ifdef HAS_NEGATIVE_Q
            // Process negative values
            plusExponentiationNeg(1.0 / term, level->powerSum);
#endif
        }

        // when monitoring Correlations, we have to keep the logs for later use
        if (monitorCorrelation) {
            double weight = weights[scale-1];
            for (int i=0; i<NUM_Q; ++i) {
                double log2y;
#ifdef NO_YLOG2X_ASM
                log2y = log2(level->powerSum[i]);
#else
                asm(
                "fld1\n"
                "fxch\n"
                "fyl2x"    // powerSumToHqConstant[scale][i] * log2(level->powerSum[i]) in one instruction
                : "=t" (log2y)
                : "0" (level->powerSum[i])
                : "st(1)"
                );
#endif
                workHQ[i] += powerSumToHqConstant[scale][i] * log2y;
                double log2yW = log2y * weight;
                sumWY[i] += log2yW;
                sumWYY[i] += log2yW * log2y;
                sumWXY[i] += log2yW * scale;
            }
            level = level->nextLevel;
            continue;
        }

        // Step 4: update the h(q) with the power sums
        for (int i=0; i<NUM_Q; ++i) {
#ifdef NO_YLOG2X_ASM
            workHQ[i] += powerSumToHqConstant[scale][i] * log2(level->powerSum[i]);
#else
            asm(
            "fyl2x\n\t"    // powerSumToHqConstant[scale][i] * log2(level->powerSum[i]) in one instruction
            "faddl %[hq]" // loads and add workHQ[i]
            : "=t" (workHQ[i])
            : "u" (powerSumToHqConstant[scale][i]), "0" (level->powerSum[i]), [hq] "m" (workHQ[i])
            : "st(1)"
            );
#endif
        }

        level = level->nextLevel;
    }

    // Error-checking code
    for (int i=0; i<NUM_Q; ++i) {
        // Use 0.0 so it can be added & subtracted later on, NaN cannot
        if (!isnormal(workHQ[i])) workHQ[i] = 0.0;
    }

    // Complete the Correlation computation only if necessary
    if (!monitorCorrelation) {
        
        // When averaging over all frames, update current sum
        if (averageAllFrames) for (int i=0; i<NUM_Q; ++i) hq[i] += workHQ[i];
        
        // h(q) computation done! the new MFA spectrum is ready :)
        return;
    }

    // for each q
    for (int i=0; i<NUM_Q; ++i) {
        // Complete the correlation coefficient computation
        double cross_product = sumWXY[i] - meanX * sumWY[i];
        workSquaredCorrelation[i] = cross_product * cross_product / (correlationCoefficientDenominatorTerm1 * (sumWYY[i] - sumWY[i] * sumWY[i] * sumOfWeightsInv));
        // Use 0.0 so it can be added & subtracted later on, NaN cannot
        if (!isnormal(workSquaredCorrelation[i])) workSquaredCorrelation[i] = 0.0;
    }

    // compute averages over frames if necessary
    if (averageAllFrames) {
        for (int i=0; i<NUM_Q; ++i) {
            workSquaredCorrelation[i] *= nframesInv;
            workHQ[i] *= nframesInv;
            squaredCorrelation[i] += workSquaredCorrelation[i];
            hq[i] += workHQ[i];
        }
    }

    // h(q) computation done! the new MFA spectrum is ready :)
}

