/***************************************************************************
 *   Copyright (C) 2005 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "DecompositionLevel.h"

#include "Analyzer.h"

DecompositionLevel::DecompositionLevel(int nlevel, int ncoefmin, Wavelet* w) :
lowPass((1 << nlevel) * ncoefmin + (w->support - 2) * ((1 << nlevel) - 1)),
highPass((1 << nlevel) * ncoefmin + (w->support - 2) * ((1 << nlevel) - 1)) {

    wavelet = w;

    // initialize power sums to 0
    powerSum = 0;

    newDataUpdateIdx = lowPass.capacity() - wavelet->support;

    // Start with the initialization version
    // Then switch to full-buffer regime when this happens
    newDataUpdatePtr = &DecompositionLevel::newDataUpdateInitPhase;

    cachedNewDataDelta = 0;
    cachedOldDataDelta = 0;

    // check if last level
    if (nlevel==0) {
        // final leaf of tree
        nextLevel = 0;
        nextLevelWaiting = 0;
        return;
    }

    // if not, create next level recursively
    nextLevel = new DecompositionLevel(nlevel-1, ncoefmin, w);
    // Create buffers in the waiting state. Actually dups of the other ones
    nextLevelWaiting = new DecompositionLevel(nlevel-1, ncoefmin, w);
}

DecompositionLevel::~DecompositionLevel() {
    if (powerSum) delete [] powerSum;
    if (nextLevel) delete nextLevel;
    if (nextLevelWaiting) delete nextLevelWaiting;
    if (cachedNewDataDelta) delete [] cachedNewDataDelta;
    if (cachedOldDataDelta) delete [] cachedOldDataDelta;
}

void DecompositionLevel::reset() {

    // clear data
    lowPass.size() = 0;
    highPass.size() = 0;

    // Start with the initialization version
    // Then switch to full-buffer regime when this happens
    newDataUpdatePtr = &DecompositionLevel::newDataUpdateInitPhase;

    // Re-initialize power sums to 0
    if (powerSum) delete [] powerSum;
    powerSum = 0;
    
    // recursive reset levels
    if (nextLevel) nextLevel->reset();
    if (nextLevelWaiting) nextLevelWaiting->reset();
}

/** Updates this buffer with new data. As this is a cyclic buffer, the oldest one is discarded.
    The waiting and active next level buffers are updated & swapped as necessary.
    This operation is constant-time O(L) with L number of decomposition levels.
    Warning: this version is designed for better performance and assumes the array is
    full (use the other method for the initialization phase).
    */
DecompositionLevel* DecompositionLevel::newDataUpdateFullRegime(double* lphp) {
    // add new data and remove the old one in one go
    lowPass.push_cyclic(lphp[0]);
    highPass.push_cyclic(lphp[1]);
    if (!nextLevel) return this;
    // Swap buffers
    DecompositionLevel* tmplp = nextLevelWaiting;
    nextLevelWaiting = nextLevel;
    nextLevel = tmplp;
    // update DWT coefficients for feeding next level recursively
    wavelet->convolve(lowPass, newDataUpdateIdx, lphp);
    return nextLevel->newDataUpdate(lphp);
}

/** Updates this buffer with new data. As this is a cyclic buffer, the oldest one is discarded.
    The waiting and active next level buffers are updated & swapped as necessary.
    This operation is constant-time O(L) with L number of decomposition levels
    This method could be used in both Initialization phase and for the full regime,
    but the other method will be more efficient for the full regime. Hence
    this function automatically switches to the other one when ready
    */
DecompositionLevel* DecompositionLevel::newDataUpdateInitPhase(double* lphp) {
    if (lowPass.isFull()) {
        // remove old data, low & high pass have same size
        lowPass.pop_front();
        highPass.pop_front();
        // Switch to full buffer version when this happens
        newDataUpdatePtr = &DecompositionLevel::newDataUpdateFullRegime;
    }
    // add new data
    lowPass.push_back(lphp[0]);
    highPass.push_back(lphp[1]);
    if (!nextLevel) return this;
    // Swap buffers
    DecompositionLevel* tmplp = nextLevelWaiting;
    nextLevelWaiting = nextLevel;
    nextLevel = tmplp;
    // update DWT coefficients for feeding next level recursively
    const int idx = lowPass.size() - wavelet->support;
    if (idx<0) return 0; // Highest level not reached yet
    wavelet->convolve(lowPass, idx, lphp);
    return nextLevel->newDataUpdate(lphp);
}

/*
void DecompositionLevel::updateScaleDelta(int dataidx, int scale, double* delta) {
    // CPU way: recompute the values each update, no need to store them
    for (int i=0; i<size; ++i) {
        delta[i] += analyzer->computeScaleDelta(dataidx, scale, highPass);
    }
}
*/
