/***************************************************************************
 *   Copyright (C) 2005 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "ScaleDeltaFilter.h"

using namespace std;

// implement global map declaration
std::map<Wavelet*, std::map<int,ScaleDeltaFilterArray*> > ScaleDeltaFilterArray::globalMap;

// factory
ScaleDeltaFilterArray* ScaleDeltaFilterArray::getInstance(Wavelet* wavelet, int nlevel) {
    std::map<Wavelet*, std::map<int,ScaleDeltaFilterArray*> >::iterator globalIt = globalMap.find(wavelet);

    // No such wavelet? => create the new filter
    if (globalIt==globalMap.end()) {
        ScaleDeltaFilterArray* filter = new ScaleDeltaFilterArray(wavelet,nlevel);
        globalMap[wavelet][nlevel] = filter;
        return filter;
    }

    // Search for this level. Not found? same as above
    std::map<int,ScaleDeltaFilterArray*>::iterator secMapIt = globalIt->second.find(nlevel);
    if (secMapIt==globalIt->second.end()) {
        ScaleDeltaFilterArray* filter = new ScaleDeltaFilterArray(wavelet,nlevel);
        globalIt->second[nlevel] = filter;
        return filter;
    }

    // Found! increase ref count
    return secMapIt->second->ref();
}

// build a new filter set with these parameters
ScaleDeltaFilterArray::ScaleDeltaFilterArray(Wavelet* wavelet, int nlevel) {

    refCount = 1;
    
    this->nlevel = nlevel;
    this->wavelet = wavelet;
    int nframes = 1 << nlevel;

    scaleDeltaFilter = new ScaleDeltaFilter[nframes];

    // Compute all filters, 2 by 2 (even and odd data)
    for (int f=0; f<nframes; f+=2) {

        scaleDeltaFilter[f].coefs = new double*[nlevel+1];
        scaleDeltaFilter[f+1].coefs = new double*[nlevel+1];
        scaleDeltaFilter[f].size = new int[nlevel+1];
        scaleDeltaFilter[f+1].size = new int[nlevel+1];
        scaleDeltaFilter[f].coefs[0] = 0;   // unused : original data is level 0
        scaleDeltaFilter[f+1].coefs[0] = 0; // unused : original data is level 0

        // Initialize
        scaleDeltaFilter[f].size[0] = 2;   // Even filter
        scaleDeltaFilter[f+1].size[0] = 2; // Odd filter

        // Local variables to ease computations
        int filterStart[nlevel+1];
        int filterEnd[nlevel+1];
        // Shift start position by enough frames to avoid negative number effects
        filterStart[0] = nframes*2 + f;
        filterEnd[0] = filterStart[0] + 1;

        // Compute fixed filter sizes for higher levels
        for (int i=1; i<=nlevel; ++i) {
            // Start position shifts by half support -1
            filterStart[i] = (filterStart[i-1] >> 1) - (wavelet->support >> 1) + 1;
            // End position remains on the same frame as the initial data
            filterEnd[i] = filterEnd[i-1] >> 1;
            // Now get the size back (remove previous shift effect)
            const int fsize = filterEnd[i] - filterStart[i] + 1;
            // Both even and odd filters have the same size
            scaleDeltaFilter[f+1].size[i] = scaleDeltaFilter[f].size[i] = fsize;
            // Allocate coefficients for this scale, they will be computed below
            scaleDeltaFilter[f].coefs[i] = new double[fsize];
            scaleDeltaFilter[f+1].coefs[i] = new double[fsize];
        }

        // tmp buffer
        double evenodd[2];

        // Process each scale in turn
        for (int s=1; s<=nlevel; ++s) {

            // The goal is to compute the filter coefficients

            // Pseudo high-pass at scale s. See below for initialization
            CyclicArray hp(scaleDeltaFilter[f].size[s]);
            hp.size() = hp.capacity(); // uninitialize

            // Thanks to linear nature of wavelet filters, we can compute the
            // final coefficients by setting the desired position to 1 and
            // the others to 0.
            for (int c = 0; c < scaleDeltaFilter[f].size[s]; ++c) {
                hp[c] = 1;
                // first time coefs are uninitialized, then only reset previous loop coef.
                if (c!=0) hp[c-1] = 0;
                else for (int i = 1; i < scaleDeltaFilter[f].size[s]; ++i) hp[i] = 0;

                // Now compute the coefficient for this position by reconstructing
                // the data from these simulated high-pass coefficients

                // reconstruction at scale-1 has this number of components
                // case s=0 is the original data, with one Even and one Odd component
                CyclicArray rec(scaleDeltaFilter[f].size[s-1]);
                rec.size() = rec.capacity(); // uninitialize

                // reconstruct down to original data scale
                // first reconstruction is delta using HP coefs
                int idx = 0;
                for (int i=0; i<rec.size(); ++i) {
                    // Idx is increased after an odd coefficient for the next even
                    wavelet->delta(hp, idx, evenodd);
                    if ((filterStart[s-1] + i) & 1) {
                        rec[i] = evenodd[1];
                        ++idx;
                    }
                    else rec[i] = evenodd[0];
                }

                // other reconstructions are smoothed versions of previous reconstruction
                for (int sr = s-1; sr>0; --sr) {
                    // We'll use previous rec for the higher scale, but we'll switch it with this array
                    // Thus, at the end of loop, previous rec will be destroyed, while current rec is
                    // kept in variable outside loop for next loop.
                    CyclicArray higher_scale(scaleDeltaFilter[f].size[sr-1]);
                    // Hehe, restore variable names to their proper meaning
                    rec.swap(higher_scale);
                    rec.size() = rec.capacity(); // uninitialize

                    // Re-use previous loop, but with smoothing.
                    idx = 0;
                    for (int i=0; i<rec.size(); ++i) {
                        // See previous comments above
                        wavelet->smooth(higher_scale, idx, evenodd);
                        if ((filterStart[sr-1] + i) & 1) {
                            rec[i] = evenodd[1];
                            ++idx;
                        }
                        else rec[i] = evenodd[0];
                    }
                }

                // rec is now guaranteed to be of size 2 => the data scale s = 0
                // Store the 2 coefficients for odd/even filters
                scaleDeltaFilter[f].coefs[s][c] = rec[0];
                scaleDeltaFilter[f+1].coefs[s][c] = rec[1];
            }
        }

    }

}

// destroy filter on last unref
ScaleDeltaFilterArray::~ScaleDeltaFilterArray() {

    // First thing to do: remove this object from the global map!
    globalMap[wavelet].erase(nlevel);
    if (globalMap[wavelet].empty()) globalMap.erase(wavelet);

    int nframes = 1 << nlevel;
    
    for (int f=0; f<nframes; f+=2) {
        for (int s=1; s<=nlevel; ++s) {
            delete [] scaleDeltaFilter[f].coefs[s];
            delete [] scaleDeltaFilter[f+1].coefs[s];
        }
        delete [] scaleDeltaFilter[f].coefs;
        delete [] scaleDeltaFilter[f+1].coefs;
        delete [] scaleDeltaFilter[f].size;
        delete [] scaleDeltaFilter[f+1].size;
    }
    delete [] scaleDeltaFilter;
}

