/***************************************************************************
 *   Copyright (C) 2005 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef SCALE_DELTA_FILTER_H
#define SCALE_DELTA_FILTER_H

#include <map>
#include "Wavelet.h"

/** This is a helper class to compute the D(s) : difference between the original
    data X, and the reconstruction R(s) without high pass coefficients up
    to scale s: D(s) = X - R(s).
    Since wavelet filters are linear combinations of coefficients, then so
    are combinations of wavelet filters. In particular, R(s) is a linear
    combination of coefficients at scale s (all high-pass coefficients of
    lower scales are removed by definition).
    Since R(s-1) and R(s) are reconstructions without HP coefficients up
    to scales s-1 and s, then the HP coefficients at scale s contain all
    the information for their difference. The same is also true for
    D(s) = X - R(s) and D(s-1) = X - R(s-1).
    Therefore, it is possible to find a formula that gives D(s) in terms
    of D(s-1) plus a linear combination (a filter) of high-pass
    coefficients at scale s: D(s) = D(s-1) + f_s(HP_s).

    The goal of this helper is to find these filters. Then, computation
    of the various D(s) can be done immediately for each scale based on previous
    results. This gives an O(L) algo instead of O(2^L) if we had to do the
    naive D(s) = X - R(s) computation for each scale.

    Note: There are 2^s possible filters for each scale, one for each data position
*/
struct ScaleDeltaFilter {
    /// double array[scale][ncoefs]
    double** coefs;

    /// int array[scale]
    int* size;
};

/// Holds the 2^L filters for direct computation of the scale delta
/// Such objects may be shared amongst different analyzers with the same wavelet/nlevel
class ScaleDeltaFilterArray {
protected:
    static std::map<Wavelet*, std::map<int,ScaleDeltaFilterArray*> > globalMap;
    
    /// Reference counting
    int refCount;

    /// protected operations. Use public factory and ref counting
    ScaleDeltaFilterArray(Wavelet* wavelet, int nlevel);
    ~ScaleDeltaFilterArray();

    /// Array of individual filters
    ScaleDeltaFilter* scaleDeltaFilter;

    /// Redundant info necessary for destructor. Too bad
    Wavelet* wavelet; int nlevel;

public:
    /// Get an instance of a filter matching this wavelet and nlevel of decompositions
    /// If no such filter exist, create it.
    /// Whenever the caller is done with using the filter, call unref()
    /// Warning: this is not a thread-safe operation
    static ScaleDeltaFilterArray* getInstance(Wavelet* wavelet, int nlevel);

    /// Ref counting functions. Warning: these are not thread-safe.
    inline ScaleDeltaFilterArray* ref() {++refCount; return this;}
    inline ScaleDeltaFilterArray* unref() {if (--refCount==0) {delete this; return 0;} return this;}

    /// Accessors to the individual filters
    inline ScaleDeltaFilter& get(const int idx) const {return scaleDeltaFilter[idx];}
    /// array syntax. same as get
    inline ScaleDeltaFilter& operator[](const int idx) const {return scaleDeltaFilter[idx];}

    /// Read-only accessor to the filter array. Caller may store this as cache
    /// to avoid the cost of one indirection in the operator[]
    inline ScaleDeltaFilter* getArray() const {return scaleDeltaFilter;}
};


#endif
