/***************************************************************************
 *   Copyright (C) 2005 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program implements the Hebbian rule experiment from the paper:   *
 *   "Learning using Dynamical Regime Identification and Synchronization"  *
 *   Presented at the WCCI06, Nicolas Brodu, 2006                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <math.h>

#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <list>
#include <map>
#include <set>
#include <stdexcept>

#include <amygdala/network.h>
#include <amygdala/inputneuron.h>
#include <amygdala/lsmtopology.h>
#include <amygdala/factory.h>
#include <amygdala/utilities.h>
#include <amygdala/logging.h>
#include <amygdala/spikeinput.h>
#include <amygdala/alphaneuron.h>
#include <amygdala/synapse.h>
#include <amygdala/basicspikeoutput.h>
#include <amygdala/statictrainer.h>
#include <amygdala/axonnode.h>
#include <amygdala/physicalproperties.h>

#include "Utility.h"
#include "Analyzer.h"

using namespace std;
using namespace Utility;
using namespace Amygdala;

// Instance is the channel values
typedef vector<double> DataInstance;
// The inputs and their classification
typedef vector< pair<DataInstance, int> > DataSet;

/// A specialized DynamicSynapse that can learn by modifying its parameters
class TrainableDynamicSynapse : public DynamicSynapse {
public:

    TrainableDynamicSynapse(DynamicSynapseProperties& synProps) : DynamicSynapse(synProps) {
    }

    virtual void Clone(void* newSynapse) {
        TrainableDynamicSynapse* s = new(newSynapse) TrainableDynamicSynapse(*this);
    }

    // delta time: efferent - afferent spikes
    // may be negative (see trainer code below)
    signed long long hebbianDelta;
    // Number of deltas that were added, for averaging purpose
    int numHebbianDelta;
    // Keep a copy of transmission delay here instead of the AxonNode to simplify further algorithms
    // See also setEfferentAfferent
    AmTimeInt transmissionDelay;

    // update connection weight
    void learn(double learningRatio) {

        // End up saturating after some time, but give a chance to increase or decrease
        A *= learningRatio;

        // Set hard limits. Another possibility would be to modify the connectivity
        if (A>=0.0) {
            if (A>90.0) A = 90.0;
            if (A<1.0) A = 1.0;
        }
        else {
            if (A<-60.0) A = -60.0;
            if (A>-1.0) A = -1.0;
        }
    }

    inline float& weight() {return A;}
};

/// A specialized Alpha neuron that maintains training stats
class TrainableAlphaNeuron : public AlphaNeuron {
friend class DynamicTrainer;
public:

    double wglob; // global classifier weight

    // define factory, propagate constructor
    typedef NeuronFactory<TrainableAlphaNeuron,AlphaNeuronProperties> Factory;
    TrainableAlphaNeuron(AmIdInt neuronId, const AlphaNeuronProperties& neuronProps) :
        AlphaNeuron(neuronId, neuronProps)
    {
        afferentSumWeights = 0.0;
        spikeCount = 0;
        wglob = Random() * 2.0 - 1.0;
    }

    double spikeCount;

    virtual ~TrainableAlphaNeuron() {}

    // Due to how Amygdala class work, each neuron is responsible for it's efferent synaptic weights only.
    // Neuron (pre-synapse) 1-> Axon N-> AxonNode N-> Synapse 1-> Dendrite 1-> Neuron (post-synapse)
    // The idea is to have a direct link to afferent neurons
    map<TrainableAlphaNeuron*, TrainableDynamicSynapse*> afferent;
    double afferentSumWeights;

    // Each node sets itself in its efferents afferent list
    void setEfferentAfferent() {

        // for each AxonNode
        for (Axon::iterator axit = GetAxon()->begin(); axit != GetAxon()->end(); ++axit) {
            // Axon node iterator syntax
            AxonNodeIterator anit = (*axit)->Begin();
            Synapse *s;
            while ((s = anit++)!=0) {
                // Connection weight corresponds to DynamicSynapse A parameter
                // Since this is protected, we have to override the class
                // Unfortunately, this imposes registering a new type of connector as well (ConnectorRegistery.register)
                // and to create LSM with that connection type
                TrainableDynamicSynapse* synapse = dynamic_cast<TrainableDynamicSynapse*>(s);
                if (!synapse) continue;

                // Sanity check for end neuron type as well
                TrainableAlphaNeuron *neuron = dynamic_cast<TrainableAlphaNeuron*>(synapse->GetDendrite()->GetPostNeuron());
                if (!neuron) continue;
                if (neuron->GetId()==GetId()) continue; // ignore self-loops

                // Reverse link
                neuron->afferent.insert(make_pair(this,synapse));
                neuron->afferentSumWeights += fabs(synapse->weight());
                // Copy the transmission delay that is in the AxonNode into this synapse
                synapse->transmissionDelay = (*axit)->GetOffset();
            }
        }
    }

// Functors & templates would be more in the C++ way than macros
// But this is just an experiment, not a software library!

#define DEFINE_GLOBAL_PROXY_PART1 \
        {std::string topName = "LSM Topology"; \
        Topology *topology = Network::GetNetworkRef()->GetTopology(topName); \
        if (!topology) throw runtime_error("TrainableAlphaNeuron:: wrong network topology!"); \
        for(Topology::iterator topit = topology->begin(); topit!=topology->end(); ++topit) { \
            TrainableAlphaNeuron* neuron = dynamic_cast<TrainableAlphaNeuron*>(*topit); \
            if (!neuron) throw runtime_error("TrainableAlphaNeuron:: wrong neuron type!"); \

#define DEFINE_GLOBAL_PROXY_PART2 \
        }}

    inline static void normalizeRecognizerWeightsGlobal() {
        int nneurons = 0;
        DEFINE_GLOBAL_PROXY_PART1
            ++nneurons;
        DEFINE_GLOBAL_PROXY_PART2
        double nneuronsInv = 1.0 / (double) nneurons;
        DEFINE_GLOBAL_PROXY_PART1
            neuron->wglob *= nneuronsInv;
        DEFINE_GLOBAL_PROXY_PART2
    }

    inline static void clearSpikeCountGlobal() {
        DEFINE_GLOBAL_PROXY_PART1
        neuron->spikeCount = 0;
        DEFINE_GLOBAL_PROXY_PART2
    }

    inline static void setEfferentAfferentGlobal() {
        DEFINE_GLOBAL_PROXY_PART1
        neuron->setEfferentAfferent();
        DEFINE_GLOBAL_PROXY_PART2
    }


    static double normalizeSpikeCountGlobal(double duration) {
        double invDur = 1.0 / duration;
        DEFINE_GLOBAL_PROXY_PART1
            neuron->spikeCount * invDur;
        DEFINE_GLOBAL_PROXY_PART2
    }

    static double computeRecognizerMatchGlobal(int& num) {
        double totalMatch = 0.0;
        num = 0;
        DEFINE_GLOBAL_PROXY_PART1
            ++num;
            totalMatch += neuron->spikeCount * neuron->wglob;
        DEFINE_GLOBAL_PROXY_PART2
        return totalMatch / (double)num;
    }

    // train the global recognizer by a simple gradient descent
    static double trainRecognizerGlobal(double target, double learningRate, bool &error) {
        int num = 0;
        double gm = computeRecognizerMatchGlobal(num);
        if (((target<0.0) && (gm<=0.0)) || ((target>0.0) && (gm>0.0))) error = false;
        else error = true;
        double term = (target - gm);
        double gradCoef = term * learningRate / num;
        // train by gradient descent
        DEFINE_GLOBAL_PROXY_PART1
            neuron->wglob += gradCoef * neuron->spikeCount;
        DEFINE_GLOBAL_PROXY_PART2
        return 0.5 * term * term;
    }

    static void displayRecognizerWeightsGlobal() {
        cout << "recognizer weights:";
        DEFINE_GLOBAL_PROXY_PART1
        cout << " n" << neuron->GetId() << ":"<<neuron->wglob;
        DEFINE_GLOBAL_PROXY_PART2
        cout << endl;
    }

    /// At neuron level we don't know and couln't care less about the current instance.
    /// This training rule is based on the hebbian modification proposal for spiking neurons in:
    ///   "Competitive Hebbian learning through spike-timing-dependent synaptic plasticity",
    ///   Sen Song, Kenneth D. Miller, L. F. Abbott. In Nature America, 2000.
    /// The difference is that we average the spiking deltas over the training set before applying the rule.
    /// This is a form of batch learning.
    /// We also apply the modification to the connection weight directly, w.r.t its current value as a ratio,
    /// instead of applying the modification w.r.t peak conductance.
    void modifyWeights(double learningRate) {

        if (afferent.empty()) return;

        // Maintain sum of weights
        double sumToNormalize = 0.0;

        for (map<TrainableAlphaNeuron*, TrainableDynamicSynapse*>::iterator it = afferent.begin(); it!=afferent.end(); ++it) {
            TrainableAlphaNeuron *aff = it->first;

            double learningRatio = 1.0;

            // process only if there were spikes
            if (it->second->numHebbianDelta>0) {
                // avg, converted to ms for the formula below
                double avgDelta = (double)it->second->hebbianDelta * 1e-3 / (double)it->second->numHebbianDelta;

                // Adapted function from [ref above], modified to use ratio
                // Use the same tau value of 20ms
                // note delta_t here is efferent - afferent, it is the opposite in the [ref above] paper
                double f;
                if (avgDelta>=0.0) {
                    // A_plus * exp(delta_t / tau_plus)
                    learningRatio = 1.0 + learningRate * exp(-0.05 * avgDelta); // 0.05 = 1.0 / 20.0
                } else {
                    learningRatio = 1.0 / (1.0 + learningRate * exp(0.05 * avgDelta)); // 0.05 = 1.0 / 20.0
                }
            }

            // apply that ratio to the synapse
            it->second->learn(learningRatio);

            sumToNormalize += fabs(it->second->weight());
        }

        // Now normalize weights, as described in the paper
        double ratio = afferentSumWeights / sumToNormalize;
        for (map<TrainableAlphaNeuron*, TrainableDynamicSynapse*>::iterator it = afferent.begin(); it!=afferent.end(); ++it) {
            it->second->weight() = (float)(ratio * it->second->weight());
        }

        return;

    }

    /// Training algo is of purely local, this global proxy calls the local function for each neuron
    static void modifyWeightsGlobal(double learningRate) {
        DEFINE_GLOBAL_PROXY_PART1
        neuron->modifyWeights(learningRate);
        DEFINE_GLOBAL_PROXY_PART2
    }

};

static TrainableAlphaNeuron::Factory MakeTrainableAlphaNeuron;


// Amygdala way of doing things
class DynamicTrainerProperties : public TrainerProperties
{
public:
    DynamicTrainerProperties() {}
    virtual ~DynamicTrainerProperties() {}
    virtual DynamicTrainerProperties* Copy() {return new DynamicTrainerProperties(*this);};
    virtual void SetProperty(const std::string&, const std::string&) {}
};

// Due to a missing check in amygdala/Dendrite.cpp, Dynamic has to inherit from Static
class DynamicTrainer : public StaticTrainer {
    virtual void Train(StaticSynapse* syn, AmTimeInt lastTransmitTime, unsigned int lastHistIdx) {}

public:
    typedef TrainerFactory<DynamicTrainer,DynamicTrainerProperties> Factory;

    DynamicTrainer(DynamicTrainerProperties& properties, std::string name ) : StaticTrainer(properties, name) {
        hebbianCutoff = 120000;
    }
    virtual ~DynamicTrainer() {}
    virtual const std::string TrainableType() const {return "Dynamic";}

    typedef list< pair<AmTimeInt, TrainableAlphaNeuron*> > RelatedList;
    RelatedList relatedSpikes;

    // Related spikes for hebbian rule
    // We must keep all spikes between now and cutoff in the past.
    // when new spike arrives, we are able to compare:
    // - if current node is efferent to past nodes: then current node reacts logically => positive hebbian
    // - efferents of that spiking neuron in the past: current node too late! => negative hebbian
    RelatedList hebbianRelatedSpikes;
    AmTimeInt hebbianCutoff;

    virtual void ReportSpike(SpikingNeuron *nrn) {
        // Note: resolution is timeStep

        TrainableAlphaNeuron* neuron = dynamic_cast<TrainableAlphaNeuron*>(nrn);
        if (neuron==0) throw runtime_error("DynamicTrainer: wrong neuron type!");

        // Maintain spike count for computing the activity signal for the global recognizer
        ++neuron->spikeCount;

        // Compare delays for the hebbian learning
        // First remove spikes after max cutoff, so as to speed up further tests
        while ((!hebbianRelatedSpikes.empty()) && (hebbianRelatedSpikes.front().first+hebbianCutoff<Network::SimTime()))
            hebbianRelatedSpikes.pop_front();

        // Maintain delta between afferent/efferent for hebbian rule:

        // If one of the past spiking neurons is in the current neuron afferent list,
        // then current node reacts logically => positive hebbian learning
        // Note: by construction, there is no duplicate in spike list
        for (RelatedList::iterator it = hebbianRelatedSpikes.begin(); it!=hebbianRelatedSpikes.end(); ++it) {
            TrainableAlphaNeuron *pastSpikingNeuron = it->second;

            // Check if past node is in afferents
            map<TrainableAlphaNeuron*, TrainableDynamicSynapse*>::iterator aff = neuron->afferent.find(pastSpikingNeuron);
            if (aff == neuron->afferent.end()) continue; // not found

            // Note: the final delta may still be negative even if the node spiked in the past, there is a transmission delay
            // However, the spikes are still received in the global time order by this function, so they are
            // in the "past" list!
            aff->second->hebbianDelta += Network::SimTime() - it->first;
            aff->second->hebbianDelta -= aff->second->transmissionDelay;
            ++aff->second->numHebbianDelta;

        }

        // efferents of that spiking neuron are in the past: current node too late! => negative hebbian learning
        // see comments in the afferent-efferent list building code for this funny loop
        // Note: by construction, there is no duplicate in the spike list
        for (Axon::iterator axit = neuron->GetAxon()->begin(); axit != neuron->GetAxon()->end(); ++axit) {
            // Non-standard iterator
            AxonNodeIterator anit = (*axit)->Begin();
            Synapse *s;
            while ((s = anit++)!=0) {
                // Sanity check for synapse type
                TrainableDynamicSynapse* synapse = dynamic_cast<TrainableDynamicSynapse*>(s);
                if (!synapse) continue;

                // Sanity check for end neuron type as well
                TrainableAlphaNeuron *eff = dynamic_cast<TrainableAlphaNeuron*>(synapse->GetDendrite()->GetPostNeuron());
                if (!eff) continue;

                // check if efferent is in past nodes list
                RelatedList::iterator found = hebbianRelatedSpikes.end();
                for (RelatedList::iterator it = hebbianRelatedSpikes.begin(); it!=hebbianRelatedSpikes.end(); ++it) {
                    if (eff->GetId() == it->second->GetId()) {found = it; break;}
                }

                // skip efferents not in past spiking list
                if (found==hebbianRelatedSpikes.end()) continue;

                // Found efferent node back in past node list!!!
                // => we're an afferent spiking after => NO!
                // subtracting positive value instead of adding negative value
                synapse->hebbianDelta -= Network::SimTime() - found->first;
                // Take the propagation delay in account, see above
                synapse->hebbianDelta += (*axit)->GetOffset();
                ++synapse->numHebbianDelta;
            }
        }

        // Add this spike to the list. An explicit check is needed for dups
        RelatedList::iterator found = hebbianRelatedSpikes.end();
        for (RelatedList::iterator it = hebbianRelatedSpikes.begin(); it!=hebbianRelatedSpikes.end(); ++it) {
            if (neuron->GetId() == it->second->GetId()) {found = it; break;}
        }
        // found a dup: remove the old spike
        if (found!=hebbianRelatedSpikes.end()) hebbianRelatedSpikes.erase(found);
        // insert the most recent one at the end
        hebbianRelatedSpikes.push_back(make_pair(Network::SimTime(), neuron));
    }

};

static DynamicTrainer::Factory MakeDynamicTrainer;

DynamicTrainer* trainer = 0;


/// Manages input spikes from either training or test set
/// Follows Amygdala conventions for factory, etc.
/// This class streams spikes from the current set it is provided
class InputsSpikeManager : public SpikeInput {
public:
    bool trainingMode;
    bool monitorTrainingError;
    ostream* monitoringTrainingFile;
    double learningRateGlobalRecognizer, learningRateRecurrentNeurons;
    int testError;
    double globRecError;

    // Number of consecutive epochs with low error after which the network is considered trained
    AmTimeInt epochsForConvergence;
    // max error level for the consecutive epochs
    double errorForConvergence;
    AmTimeInt convergedEpochs;

    typedef SpikeInputFactory<InputsSpikeManager> Factory;

    InputsSpikeManager(const std::string& name) :
        SpikeInput(name),
        dataSet(0),
        currentSpikeTrainClassification(0),
        duration(0), currentBaseTime(0), nepoch(0), convergedEpochs(0),
        monitoringTrainingFile(0), epochsForConvergence(50), errorForConvergence(1e-6),
        trainingMode(true), monitorTrainingError(true), finalizingEpoch(false)
    {}

    InputsSpikeManager() :
        SpikeInput(),
        dataSet(0),
        currentSpikeTrainClassification(0),
        duration(0), currentBaseTime(0), nepoch(0), convergedEpochs(0),
        monitoringTrainingFile(0), epochsForConvergence(50), errorForConvergence(1e-6),
        trainingMode(true), monitorTrainingError(true), finalizingEpoch(false)
    {}

    virtual ~InputsSpikeManager() {}

    typedef struct {double currentLevel, currentInput, baseInput, sensitivity, threshold;} ReceptorParam;
    vector<ReceptorParam> receptors;

    inline int createReceptorChannels(int nchannels, int numReceptorsPerChannel, double minFreq, double maxFreq, double variance, double threshold) {

        // Init receptor channel
        // This is done before adding neurons to this neuron group
        // return number of necessary neuron as convenience
        // This will also be our future neuron group size.
        int nneurons = nchannels * numReceptorsPerChannel;
        receptorChannel.resize(nneurons);
        // Put all neurons channel after channel, numReceptorsPerChannel times
        for (int i=0; i<nneurons; ++i) {
            receptorChannel[i] = i % nchannels;
        }
        // Amygdala creates neurons side by side geographically
        // mixup receptors for the != input channels
        for (int i=0; i<nneurons; ++i) {
            int i1 = IRandom(0, nneurons-1);
            int i2 = IRandom(0, nneurons-1);
            int tmp = receptorChannel[i1];
            receptorChannel[i1] = receptorChannel[i2];
            receptorChannel[i2] = tmp;
        }

        // Each receptor has slightly different parameters
        // Simple modelling: accumulate=integrate input over time, then spike when the threshold is reached
        // In case of constant inputs, this corresponds to a frequency response linear w.r.t. the input value
        // 3 parameters: . base Input for making input=0 spike eventually.
        //               . threshold = potential to reach to spike. Arbitrary, should be consistent with LSM nA scaling
        //               . receptor sensitivity = coef for input.
        // receptors start with current level = 0, it's reset to 0 after spike
        // A perhaps better modelling would consider a full alpha neuron with refactory period, etc.
        receptors.resize(nneurons);

        // wavelenghts in timeStep units, for correct integration
        double maxL = 1e6 / minFreq /* now in usec */  /  (double)Network::TimeStepSize(); // now in timestep units
        double minL = 1e6 / maxFreq /* now in usec */  /  (double)Network::TimeStepSize(); // now in timestep units
        // base baseInput
        double b = threshold / maxL;
        // base sensitivity
        double a = b * (maxL - minL) / minL;

        for (int i=0; i<nneurons; ++i) {
            receptors[i].currentLevel = 0.0;
            receptors[i].currentInput = 0.0;
            receptors[i].baseInput = b * (1.0 + NRandom()*variance);
            receptors[i].sensitivity = a * (1.0 + NRandom()*variance);
            receptors[i].threshold = threshold * (1.0 + NRandom()*variance);
        }

        return nneurons;
    }

    /// One may double-check rounding errors didn't wreck the duration badly
    inline double getSpikeTrainDuration() const {
        return duration * 1e-6;
    }

    // Flag that indicate if we're processing the last epoch or not (to pre-load the next)
    bool finalizingEpoch;

    /// Use this to switch between training and testing sets
    inline void setDataSet(DataSet& dataSet, double duration, int epoch=0) {
        this->dataSet = &dataSet;
        // internally store duration as number of usec
        this->duration = static_cast<AmTimeInt>(duration * 1e6);
        finalizingEpoch = false;

        // Cleanup previous parameters
        permutation.clear();
        nepoch = epoch; testError = 0; convergedEpochs = 0; globRecError = 0.0;
        // randomize set order
        makePerm();
        // Pre-fill first spikes
        fillSpikes();
    }

    /// Use this for training and error computations: this is the real(=target) data classification
    inline int getCurrentSpikeTrainClassification() const {
        return currentSpikeTrainClassification;
    }

    /// Read input for that timeStep
    /// Implements parent pure virtual streaming function
    virtual void ReadInputBuffer() {

        // Called each time step => perform integration for receptors

        AmTimeInt simTime = Network::SimTime();
        Network* theNet = Network::GetNetworkRef();

        if (simTime==Network::GetNetworkRef()->GetMaxRunTime()) finalizingEpoch=true;

        for (int i=0; i<receptors.size(); ++i) {
            // integrate for this time step
            receptors[i].currentLevel += receptors[i].currentInput * receptors[i].sensitivity + receptors[i].baseInput;
            if (receptors[i].currentLevel >= receptors[i].threshold) {
                // schedule spike & reset if trigger happens
                theNet->ScheduleSpike(simTime, static_cast<InputNeuron*>(GetNeuron(i)));
                receptors[i].currentLevel = 0.0;
            }
        }

        // check for new data
        if (simTime >= currentBaseTime + duration) {
            // Auto post-processing
            postProcessPreviousSpikeTrain();
            if (permutation.empty()) {
                postProcessPreviousSpikeSet();
                if (!finalizingEpoch) makePerm(); // prepare next epoch if necessary
            }
            // Prepare next spike train
            currentBaseTime += duration;
            if (!finalizingEpoch) fillSpikes();
        }

    }

    void postProcessPreviousSpikeTrain() {

        // Post-processing differs between training and testing modes!
        if (trainingMode) {
            TrainableAlphaNeuron::normalizeSpikeCountGlobal(duration);
            bool hasMadeAnError;
            double err = TrainableAlphaNeuron::trainRecognizerGlobal(currentSpikeTrainClassification*2.-1., learningRateGlobalRecognizer, hasMadeAnError);
            cout << "Error = " << err << endl;
            globRecError += err;
            if (hasMadeAnError) ++testError;

            TrainableAlphaNeuron::modifyWeightsGlobal(learningRateRecurrentNeurons);

        // testing just computes the classification error
        } else {
            int num;
            TrainableAlphaNeuron::normalizeSpikeCountGlobal(duration);
            double gm = TrainableAlphaNeuron::computeRecognizerMatchGlobal(num);
            double estimatedClass = (gm>0.0) ? 1 : 0;
            if (currentSpikeTrainClassification!=estimatedClass) ++testError;
        }

        TrainableAlphaNeuron::clearSpikeCountGlobal();
    }


    // Only for display & log facilities
    void postProcessPreviousSpikeSet() {
        if (trainingMode) cout << "Global recognizer error: " << (globRecError / (double)dataSet->size()) << endl;
        cout << "Classification error: " << (testError / (double)dataSet->size()) << endl;
        if (monitoringTrainingFile!=0) (*monitoringTrainingFile) << nepoch << " " << (globRecError / (double)dataSet->size()) << " " << (testError / (double)dataSet->size()) << endl;
    }

    AmTimeInt currentBaseTime;

    // size of this vector in ninputs, for each input neuron it associates the channel = index in data vector
    vector<int> receptorChannel;

    void fillSpikes() {

        // Get a random spike train.
        if (permutation.empty()) throw runtime_error("InputsSpikeManager::fillSpikes: no permutation!");
        int r = permutation.back(); permutation.pop_back();

        // Updates current reference instance number
        currentSpikeTrainClassification = (*dataSet)[r].second;

        cout << "Epoch " << nepoch << ", Instance " << (r+1) << ", " << (dataSet->size() - permutation.size()) << "/" << dataSet->size() << " (" << currentSpikeTrainClassification << ")" <<endl;

        const DataInstance& data = (*dataSet)[r].first;

        // Population coding. Set corresponding channel new input to all receptors
        for (int i=0; i<receptors.size(); ++i) {
            receptors[i].currentInput = data[receptorChannel[i]];
        }
    }

    vector<int> permutation;
    int nepoch;

    /// Makes a permutation to run each instance of the set once, in random order
    void makePerm() {

        cout << endl << "Epoch " << (++nepoch) << endl << endl;

        globRecError = 0.0;
        testError = 0;

        // prepare permutation for the correct number of indices
        permutation.resize(dataSet->size());
        // first assign all indices
        for (int i=0; i<permutation.size(); ++i) permutation[i] = i;
        // then shuffle
        for (int i=0; i<permutation.size(); ++i) {
            int i1 = IRandom(0, permutation.size()-1);
            int i2 = IRandom(0, permutation.size()-1);
            int tmp = permutation[i1];
            permutation[i1] = permutation[i2];
            permutation[i2] = tmp;
        }
    }

    // Cannot store reference due to factories
    DataSet* dataSet;
    int currentSpikeTrainClassification;
    AmTimeInt duration;
};

// Amygdala way of doing things
static InputsSpikeManager::Factory MakeInputsSpikeManager;

// callback to control RNG in amygdala with our seed
float AmygdalaRNG() {
    return (float)Random();
}

// Reads a parameter line in the form "xxx=yyy"
inline int readProben1Parameter(ifstream& dataFile) {
    string line;
    int param;
    dataFile >> line;
    dataFile.get(); // \n
    stringstream converter(line.substr(line.find("=")+1,string::npos));
    converter >> param;
    return param;
}

int main(int argc, char *argv[])
{
    long seed = 0;
    if (argc>1) {
        seed = atol(argv[1]); // 0 in case of failure
        if (!seed) cout << "Warning: given seed " << argv[1] << " is invalid. Using a random one." << endl;
    }
    if (seed!=0) RandomInit(seed);
    else seed = RandomInit();

    // Sets a custom RNG instead of rand()/RAND_MAX
    // Allows to reproduce experiments with our seed
    Utilities::setRNG(&AmygdalaRNG);

    int npinputs, npoutputs, ntraining, ntest, nchannels;
    int numReceptorsPerChannel = 5; // Amygdala throws runtime_error if not enough space
    int LiquidX = 6, LiquidY = 6, LiquidZ = 6;
    ifstream dataFile;

    if (argc>2) {
        // Using proben1 file format for data set

        // read proben1 arguments in npxxx variables
        dataFile.open(argv[2]);
        if (!dataFile) {cerr << "Can't read Proben1 datafile: " << argv[2] << endl; return 1;}

        cout << "Using data file " << argv[2] << endl;

        // For our purpose, bool and real inputs are the same.
        // bool_in=X
        npinputs = readProben1Parameter(dataFile);
        // real_in=X
        npinputs += readProben1Parameter(dataFile);

        // Outputs: . note when dealing with booleans, because the search space is then finite
        //          and an exhaustive search may be possible
        //          . general case: consider real and booleans as numbers, like for the inputs
        // bool_out=X
        npoutputs = readProben1Parameter(dataFile);
        if (npoutputs!=2) {cerr << "Only binary outputs are currently supported" << endl; return 2;}
        // real_out=X
        if (readProben1Parameter(dataFile)!=0) {cerr << "Only boolean outputs are currently supported" << endl; return 2;}

        // This algorithm does not use a validation set. As recommended in the Proben1 paper,
        // both training and validation sets are thus merged.
        // training_examples=XXX
        ntraining = readProben1Parameter(dataFile);
        // validation_examples=XXX
        ntraining += readProben1Parameter(dataFile);

        // Test set: It MUST NOT be used during training. At all.
        // test_examples=XXX
        ntest = readProben1Parameter(dataFile);

        // Sanity check
        if ((npinputs==0) || (npoutputs==0) || (ntraining==0) || (ntest==0)) {
            cerr << "Error in data file header." << endl;
            return 3;
        }

        // Further lines contain the inputs & outputs
        nchannels = npinputs;
    }
    else {
        cout << "Using built-in test." << endl;
        ntraining = 20, ntest = 20;
        nchannels = 10;
    }

    // Declare data sets at global scope
    DataSet trainingInstances(ntraining), testInstances(ntest);

    if (argc>2) {
        // Process each line in turn
        // It would be faster to just >> all values as double, but the way it's done here
        // clearly separates the logical lines, training, and test sets
        int lineNum = 0;
        while (dataFile.good()) {
            // training and test instances are read the same way
            pair<DataInstance, int> &entry = (lineNum<ntraining) ? trainingInstances[lineNum] : testInstances[lineNum-ntraining];
            DataInstance& instance = entry.first;

            // Read file line by line
            stringstream ss;
            dataFile.get(*ss.rdbuf());

            // Read all input values on that line
            for (int i=0; i<npinputs; ++i) {
                double value;
                ss >> value;
                // proben inputs in 0..1 => OK
                instance.push_back(value);
            }

            // convert output into class
            entry.second = -1; // flag
            for (int i=0; i<npoutputs; ++i) {
                double value;
                ss >> value;
                // find correct class
                if (value>0.5) {
                    if (entry.second!=-1) {cerr << "More than one class declared, data line " << (lineNum+1) << endl; return 1;}
                    entry.second = i;
                }
            }
            if (entry.second==-1) {cerr << "No class declared in output, data line " << (lineNum+1) << endl; return 1;}
            // Prepare for next loop
            ++lineNum;
            if (lineNum>=ntraining+ntest) break;
            dataFile.get(); // \n
        }

        // Sanity check
        if ((trainingInstances.size()!=ntraining) || (testInstances.size()!=ntest)) {
            cerr << "Error in data file." << endl;
            return 4;
        }

    }

    // Built-in test
    else {

        // Generate the data set.
        map<int, int> numInstances; numInstances[0] = numInstances[1] = 0;

        vector<double> templates[2];
        for (int i=0; i<nchannels; ++i) templates[0].push_back(Random());
        for (int i=0; i<nchannels; ++i) templates[1].push_back(Random());

        double jitter = 4e-3;

        for (int i=0; i<ntraining+ntest; ++i) {

            // training and test instances are generated the same way
            pair<DataInstance, int> &entry = (i<ntraining) ? trainingInstances[i] : testInstances[i-ntraining];
            DataInstance& instance = entry.first;

            // Choose a classification
            int numTemplate;
            if (i<(ntraining/2)) numTemplate = 0;  // half training are 0
            else if (i<ntraining) numTemplate = 1; // half training are 1
            else numTemplate = IRandom(0,1);       // testing are random
            entry.second = numTemplate;

            // Generate base data
            instance = templates[numTemplate];

            if (i<ntraining) ++numInstances[numTemplate];

            // Add jitter. Suggestion from comment says use normal distribution
            for (int channel=0; channel<instance.size(); ++channel) {
                instance[channel] += NRandom() * jitter;
                // bounds checking
                if (instance[channel]<0.0) instance[channel] = 0.0;
                if (instance[channel]>1.0) instance[channel] = 1.0;
            }
        }
    }

    Network::Init(argc, argv);
    try{

        // This object is responsible for creating neurons & their connections
        // Uses the "Liquid State Machine" model
        LSMTopology *topology = Factory::MakeLSMTopology("LSM Topology");

        // Parameters to create neurons
        TrainableAlphaNeuron::Factory neuronFactory;
        AlphaNeuronProperties neuronProperties(true);
        neuronProperties.SetMembraneConst(30.0);
        neuronProperties.SetExcitatorySynapseConst(3.0);
        neuronProperties.SetInhibitorySynapseConst(6.0);

        // Create the main neurons according to this topology
        topology->MakeCube(&neuronFactory, &neuronProperties, LiquidX, LiquidY, LiquidZ);

        // Adds a trainer to the topology
        DynamicTrainerProperties dynamicTrainerProperties; // unused for now
        trainer = MakeDynamicTrainer(dynamicTrainerProperties, topology, "Trainer");

        // This is the input manager.
        // It is responsible for sending inputs to the input neurons
        InputsSpikeManager* inputManager = MakeInputsSpikeManager("Input Manager");

        // Maximum number of epochs to train the network
        AmTimeInt nepochs = 50;
        // Number of consecutive epochs with low error after which the network is considered trained
        AmTimeInt epochsForConvergence = 10;
        // max error level for the consecutive epochs
        double errorForConvergence = 1e-6;

        // Set a learning rate for the global recognizer and the recurrent neurons
        inputManager->learningRateGlobalRecognizer = 0.01;
        inputManager->learningRateRecurrentNeurons = 0.1;

        double duration = 1.0;

        // Tells the topology to add input neurons receiving instruction from that manager
        int ninputs = inputManager->createReceptorChannels(
            nchannels,              // int nchannels
            numReceptorsPerChannel, // int numReceptorsPerChannel
            1.0,                    // double minFreq
            50.0,                   // double maxFreq
            0.1,                    // double variance
            50.0                    // double threshold (arbitrary, no effect)
        );
        topology->MakeInput(inputManager, ninputs);

        // Synapse values for all possible combinations of inhibitory / excitatory
        // values from Maass et al. "A Model for Real Time Computation in Neural Microcircuits"
        // Four values are EE, EI, IE, II for Excitatory/Inhibitory connections
        const LSMTopology::EI A = {30., 60., -19, -19};      // scaling param in nA
        const LSMTopology::EI delay = {1500, 800, 800, 800};
        const LSMTopology::EI C = {0.3, 0.2, 0.4, 0.1};      // likelyhood of connection
        const LSMTopology::EI U = {0.5, 0.05, 0.25, 0.32};   // utilization as fraction of 1
        const LSMTopology::EI D = {1.1, 0.125, 0.7, 0.144};  // time const in sec
        const LSMTopology::EI F = {0.05, 1.2, 0.02, 0.06};   // time const in sec

        // Need patched Amygdala, 1.4.1 code has non-functional Register function
        ConnectorRegistry::GetRegistry().Register("TrainableDynamicSynapse", new NeuronConnector<TrainableDynamicSynapse, DynamicSynapseProperties>);
        topology->Connect(2.0, A, delay, C, U, D, F, "TrainableDynamicSynapse");

        // OutputManager manages program outputs (as spikes), not output neurons!
        OutputManager::EnableOutput(true, LSMTopology::INPUTGROUP);
        OutputManager::EnableOutput(true, LSMTopology::HIDDENGROUP);

        // Make log file. Basic one here, could make a stats one
        //Factory::MakeBasicSpikeOutput("output")->LogSpikeTimes("lsmspikes.log", 0, maxRunTime);
        stringstream ss;
        ss << "trainingError_hebbian_" << (argc>2?"data_":"") << seed << ".txt";
        ofstream tlog(ss.str().c_str());
        inputManager->monitoringTrainingFile = &tlog;
        inputManager->epochsForConvergence = epochsForConvergence;
        inputManager->errorForConvergence = errorForConvergence;

        AmTimeInt maxRunTime = nepochs * ntraining * static_cast<AmTimeInt>(duration * 1e6);
        cout << "maxRunTime="<<maxRunTime<<endl;

        // Starts with the training set
        inputManager->setDataSet(trainingInstances, duration);

        // Preparation for the main loop
        Network::GetNetworkRef()->InitRun();

        // Reverse-link to have direct access to afferent nodes
        TrainableAlphaNeuron::setEfferentAfferentGlobal();

        // Normalize global recognizer weights
        TrainableAlphaNeuron::normalizeRecognizerWeightsGlobal();

        // Run for nepochs
        Network::GetNetworkRef()->Run(maxRunTime);

        // update in case of premature stopping
        maxRunTime = Network::GetNetworkRef()->GetMaxRunTime();

        // Cleanup for the training phase
        tlog.close();

        map<int, double> totalMap; totalMap[0] = totalMap[1] = 0.0;


// Phase 2: measuring errors

        cout << endl << "Computing error on the test set" << endl << endl;

        // log testing
        stringstream ss2;
        ss2 << "testingError_hebbian_" << (argc>2?"data_":"") << seed << ".txt";
        tlog.open(ss2.str().c_str());

        // Proceed to the test error
        inputManager->trainingMode = false;

        // Stat holders
        double meanTest = 0.0;
        double varTest = 0.0;

        // Average over 10 epochs to make stats
        int nstat = 10;

        // Process each epoch one by one so as to maintain stats
        for (int statLoopCpt = 0; statLoopCpt<nstat; ++statLoopCpt) {
            inputManager->setDataSet(testInstances, duration, statLoopCpt);
            nepochs = 1;
            maxRunTime += nepochs * ntest * static_cast<AmTimeInt>(duration * 1e6);
            Network::GetNetworkRef()->Run(maxRunTime);
            double sampleError = (double)inputManager->testError / (double)ntest;
            cout << "Stat loop " << (statLoopCpt+1) << "/" << nstat << ": sample testing error: " << sampleError << endl;

            // maintain stats
            meanTest += sampleError;
            varTest += sampleError * sampleError;
        }

        // finish stat computations
        meanTest /= (double)nstat;
        varTest = varTest / (double)nstat - meanTest * meanTest;
        if (varTest<0.0) varTest = 0.0; // numerical roundup may occur

        cout << "Overall testing error: " << meanTest << ", std. dev. = " << sqrt(varTest) << endl;
        tlog << endl << "Overall testing error (mean std. dev.): " << endl << meanTest << " " << sqrt(varTest) << endl;

        //Network::GetNetworkRef()->Save("network.xml.gz");

    } catch (runtime_error e){
        cerr << "Caught Runtime Error: " << e.what() << endl;
        cerr << "Exiting." << endl;
        exit(-1);
    }
    return 0;

}

