/***************************************************************************
                          alphaneuron.cpp  -  description
                             -------------------
    copyright            : (C) 2001 by Matt Grover
    email                : mgrover@amygdala.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <cmath>

#include "alphaneuron.h"
#include "euler.h"
#include "network.h"
#include "functionlookup.h"
#include "dendrite.h"
#include "logging.h"
#include "utilities.h"
#include <iostream>

using namespace std;
using namespace Amygdala;

unsigned int AlphaNeuron::pspStepSize = 0;
unsigned int AlphaNeuron::pspLSize = 0;

/////////////////////////////////////////////////////
//// AlphaNeuronProperties
/////////////////////////////////////////////////////
AlphaNeuronProperties::AlphaNeuronProperties():
	SpikingNeuronProperties()
{}

AlphaNeuronProperties::AlphaNeuronProperties(bool initializePhysicalProps):
	SpikingNeuronProperties(initializePhysicalProps)
{}

AlphaNeuronProperties::AlphaNeuronProperties(const AlphaNeuronProperties& rhs):
	SpikingNeuronProperties(rhs)
{
	memTimeConst = rhs.memTimeConst;
	eSynTimeConst = rhs.eSynTimeConst;
	iSynTimeConst = rhs.iSynTimeConst;
}

AlphaNeuronProperties::~AlphaNeuronProperties()
{}

AlphaNeuronProperties* AlphaNeuronProperties::Copy() const
{
	return new AlphaNeuronProperties(*this);
}

void AlphaNeuronProperties::SetProperty(string& name, string& value)
{
    // Properties:
    // memTimeConst
    // eSynTimeConst
    // iSynTimeConst
    if (!name.compare("membraneTimeConst")) {
        memTimeConst = atof(value.c_str());
    }
    else if (!name.compare("eSynapticTimeConst")) {
        eSynTimeConst = atof(value.c_str());
    }
    else if (!name.compare("iSynapticTimeConst")) {
        iSynTimeConst = atof(value.c_str());
    }
    else {
        NeuronProperties::SetProperty(name, value);
    }
}

map< string, string > AlphaNeuronProperties::GetPropertyMap() const
{
    map< string, string > propMap = NeuronProperties::GetPropertyMap();

    propMap["membraneTimeConst"] = Utilities::ftostr(memTimeConst);
    propMap["eSynapticTimeConst"] = Utilities::ftostr(eSynTimeConst);
    propMap["iSynapticTimeConst"] = Utilities::ftostr(iSynTimeConst);
    
    return propMap;
}

/////////////////////////////////////////////////////
//// AlphaNeuron
/////////////////////////////////////////////////////
AlphaNeuron::AlphaNeuron(AmIdInt neuronId, const AlphaNeuronProperties& neuronProps):
    SpikingNeuron(neuronId, neuronProps),
    histBeginIdx(0),
    maxThreshCrs(0),
    convergeRes(0)
{
    epspLookup = 0;
    edPspLookup = 0;
    ipspLookup = 0;
    idPspLookup = 0;
    
    if (!pspStepSize) {
	    Init();
    }
    
    InitLookup();
}

AlphaNeuron::~AlphaNeuron()
{
}

void AlphaNeuron::Init()
{
	//pspStepSize = Network::TimeStepSize()/10;
	pspStepSize = Network::TimeStepSize();
	pspLSize = 100000/pspStepSize;		// 100ms / step size
	
	LOGGER(3, "pspStepSize: " << pspStepSize << " pspLSize: " << pspLSize)
}

void AlphaNeuron::SetProperties(AlphaNeuronProperties* props)
{
	delete neuronProps;
	neuronProps = props->Copy();
}

void AlphaNeuron::SpikeCleanup()
{
	// reset inputHist vector
	inputHist.clear();
	histBeginIdx = 0;
}

void AlphaNeuron::ProcessInput(const AmTimeInt& inTime)
{
    unsigned int i, iterate, converged, histSize, tblIndex;
    AmTimeInt calcTime, funcTime;
    double currState = 0.0;
    double currDeriv = 0.0;
    double funcWeight = 0.0;
    double stateDelta = 0.0;
    double threshCrs = 0.0;
    double lstThreshCrs = 0.0;
    InputHist tmpInput;

    // If the neuron is within a refractory period,
    if ( (inTime - spikeTime) <= refPeriod ) {
        if (inTime > refPeriod) {
	        dendrite->ResetTrigger();
            return;
        }
    }

    iterate = 1;
    converged = 0;
    calcTime = 0;
    funcTime = 0;
    histSize = 0;

    if (!maxThreshCrs) {
        maxThreshCrs = pspLSize * pspStepSize;
        LOGGER(5, "maxThreshCrs = " << maxThreshCrs)

        // find the convergence resolution (the resolution at which two values
        // of threshCrs are considered to be identical) -- must be <= simStepSize
        if (simStepSize > pspStepSize) {
            if (pspStepSize > (simStepSize / 2.0)) {
                convergeRes = pspStepSize;
            }
            else {
                convergeRes = simStepSize / 2.0;
            }
        }
        else {
            convergeRes = simStepSize;
        }
    }

    calcTime = inTime;
    inputTime = inTime;
    currTime = inTime;
    float inWeightPos, inWeightNeg;

    dendrite->GetStimulationLevel(inWeightPos, inWeightNeg);
    /*if (trainingMode) {
        //inWeight = inputHeader->SumQueue(true, synapseHist, inTime);
        dendrite->GetStimulationLevel(inWeightPos, inWeightNeg);
    }
    else {
        //inWeight = inputHeader->SumQueue(true);
        dendrite->GetStimulationLevel(inWeightPos, inWeightNeg);
    }*/

    tmpInput.time = inTime;
    if (inWeightPos) {
        tmpInput.weight = inWeightPos;
        inputHist.push_back(tmpInput);
    }
    if (inWeightNeg) {
        tmpInput.weight = inWeightNeg;
        inputHist.push_back(tmpInput);
    }
    
    histSize = inputHist.size();

    //LOG_WEIGHT_HIST(6)
    
    i = 0;

    // Use Newton's method to determine if and when the spike will occur
    /**************************************************************************
    *    1) Determine the membrane potential (currState) at time (calcTime -
    *       inTimeHist[i]) by summing the state of each inTimeHist[]
    *       (use pspLookup).
    *    2) Find the derivative of the function for calcTime (dPspLookup).
    *    3) Calculate intercept with thresholdPtnl.
    *    4) Set new calcTime to time of intercept.
    *    5) Repeat until:
    *        a) Two successive iterations result in no change in calcTime.
    *            (Converges)
    *        b) The derivative of the function becomes negative.
    *            (Does not converge)
    **************************************************************************/
    LOGGER(6, "NEURON " << nId << " Starting main loop...")

    lstThreshCrs = static_cast<double>(calcTime);
    while (iterate) {
        currState = 0.0;
        currDeriv = 0.0;

        Utilities::RoundTime(calcTime, pspStepSize);

        LOGGER(6, "calcTime: " << calcTime)

        for (i=histBeginIdx; i<histSize; i++) {
            tmpInput = inputHist[i];
            funcTime = calcTime - tmpInput.time;
            funcWeight = tmpInput.weight;

            LOGGER(6, "funcTime: " << funcTime << "\n\tfuncWeight: " << funcWeight)

            tblIndex = (funcTime / pspStepSize);
            LOGGER(6, "tblIndex for psp lookups: " << tblIndex)
            if (tblIndex < pspLSize) {
                if ( funcWeight > 0.0 ) {
                    currState = currState + (funcWeight * (epspLookup[tblIndex]));
                    currDeriv = currDeriv + (funcWeight * (edPspLookup[tblIndex]));
                }
                else {
                    currState = currState + (funcWeight * (ipspLookup[tblIndex]));
                    currDeriv = currDeriv + (funcWeight * (idPspLookup[tblIndex]));
                }
            }
            else {
                if (calcTime <= inTime)
                    ++histBeginIdx;
            }
        }

        LOGGER(6, "currState: " << currState << "\n\tcurrDerive: " << currDeriv)

        if ( (currDeriv < 0.0) && (currState < 1.0) ) {
            converged = 0;
            iterate = 0;

            // set the scheduled spike time to zero to keep
            // the neuron from spiking when an inhibitory input
            // spike has canceled a previously scheduled spike.
            schedSpikeTime = 0;
        }
        else if (currState > 1.0) {
            converged = 1;
            iterate = 0;
        }
        else {
            stateDelta = fabs(1.0 - currState);
            threshCrs = (stateDelta / currDeriv) + lstThreshCrs;
            LOGGER(6, "stateDelta: " << stateDelta << "\nthreshCrs: " << threshCrs << "\nlstThreshCrs: " << lstThreshCrs)
            if ((threshCrs - lstThreshCrs) < convergeRes) {
                if (stateDelta < 1.0) {
                    converged = 1;
                    iterate = 0;
                    LOGGER(6, "threshCrs - lstThreshCrs < convergeRes -- Converged")
                }
                else {
                    converged = 0;
                    AmTimeInt newCalcTime = static_cast<AmTimeInt>(threshCrs);
                    if (newCalcTime == calcTime) iterate=0;
                    calcTime = newCalcTime;
                    lstThreshCrs = threshCrs;

                    // set the scheduled spike time to zero to keep
                    // the neuron from spiking when an inhibitory input
                    // spike has canceled a previously scheduled spike.
                    schedSpikeTime = 0;
                }
            }
            else {
                converged = 0;
                if ((threshCrs-lstThreshCrs) > maxThreshCrs) {
                //if (threshCrs > maxThreshCrs) {
                    iterate = 0;
                }
                AmTimeInt newCalcTime = static_cast<AmTimeInt>(threshCrs);
                if (newCalcTime == calcTime) iterate=0;
                calcTime = newCalcTime;
                lstThreshCrs = threshCrs;

                // set the scheduled spike time to zero to keep
                // the neuron from spiking when an inhibitory input
                // spike has canceled a previously scheduled spike.
                schedSpikeTime = 0;
            }
        }
    }
    if (converged) {

        // Make sure we wait at least one cycle before spiking
        if (calcTime == inTime) {
            calcTime += simStepSize;
        }
        schedSpikeTime = calcTime;

        LOGGER(5, "Neuron " << nId << " Scheduling spike at " << schedSpikeTime)
        
		Network::GetNetworkRef()->ScheduleSpike( schedSpikeTime, this );
    }
}

float AlphaNeuron::GetMembranePtnl() const
{
    float currState=0.0, currDeriv=0.0;
    unsigned int histSize = inputHist.size();
    unsigned int hbi = histBeginIdx;
    //AmTimeInt calcTime = Node::SimTime();
    AmTimeInt calcTime = Network::GetNetworkRef()->SimTime();

    Utilities::RoundTime(calcTime, pspStepSize);

    for (unsigned int i=hbi; i<histSize; i++) {
        InputHist tmpInput = inputHist[i];
        AmTimeInt funcTime = calcTime - tmpInput.time;
        float funcWeight = tmpInput.weight;

        unsigned int tblIndex = (funcTime / pspStepSize);
        if (tblIndex < pspLSize) {
            if ( funcWeight > 0.0 ) {
                currState = currState + (funcWeight * (epspLookup[tblIndex]));
                currDeriv = currDeriv + (funcWeight * (edPspLookup[tblIndex]));
            }
            else {
                currState = currState + (funcWeight * (ipspLookup[tblIndex]));
                currDeriv = currDeriv + (funcWeight * (idPspLookup[tblIndex]));
            }
        }
        else {
            ++hbi;
        }
    }
    return currState;
}

void AlphaNeuron::ScaleFunctionLookups()
{
    unsigned int i;
    float lastPtnl=0.0;
    float maxPtnl=0.0;

    for (i=0; i<pspLSize; ++i) {
        maxPtnl = epspLookup[i];
        //maxDPtnl = edPspLookup[i];

        if (maxPtnl < lastPtnl) {
            maxPtnl = lastPtnl;
            break;
        }
        else {
            lastPtnl = maxPtnl;
        }

        /*if (maxDPtnl < lastDPtnl) {
            maxDPtnl = lastDPtnl;
            break;
        }
        else {
            lastDPtnl = maxDPtnl;
        } */
    }

    if (maxPtnl == 0.0)
        return;

    // Normalize the function -- a small number is added to the
    // peak to make sure that a weight of 1 will cross the threshold
    float scaleFactor = 1.0/maxPtnl + 0.000001;
    //float dscaleFactor = 1.0/maxDPtnl;
    for (i=0; i<pspLSize; i++) {
        epspLookup[i] = scaleFactor*epspLookup[i];
        edPspLookup[i] = scaleFactor*edPspLookup[i];
    }

    // maxScaledWeight is increased a little bit beyond its
    // true value to make sure that a weight of 1.0 will always
    // exceed the threshold at the peak of the psp curve.
//    maxScaledWeight = ( thresholdPtnl / maxPtnl ) + 0.000001;

    // Scale the inhibitory psp to match the amplitude of the epsp.
    // This will ensure consistent behavior for all time constants
    // (the maximum of the ipsp gets lower with increasing synaptic
    // time constants).
    float imaxPtnl = 0.0;
    lastPtnl = 0.0;
    for (i=0; i<pspLSize; i++) {
        imaxPtnl = ipspLookup[i];
        //imaxDPtnl = idPspLookup[i];

        if (imaxPtnl < lastPtnl) {
            imaxPtnl = lastPtnl;
            break;
        }
        else {
            lastPtnl = imaxPtnl;
        }

        /*if (imaxDPtnl < lastDPtnl) {
            imaxDPtnl = lastDPtnl;
            break;
        }
        else {
            lastDPtnl = imaxDPtnl;
        }*/
    }

    //scaleFactor = 1.0*maxPtnl/imaxPtnl;
    scaleFactor = 1.0/imaxPtnl;
    //dscaleFactor = 0.9*maxDPtnl/imaxDPtnl;
    for (i=0; i<pspLSize; i++) {
        ipspLookup[i] = scaleFactor*ipspLookup[i];
        idPspLookup[i] = scaleFactor*idPspLookup[i];
    }
}

void AlphaNeuron::PrintLookupTables()
{
    cout << "\n\nPrinting excitatory tables:\nepspLookup:\n\n";
    for (unsigned int i=0; i<pspLSize; ++i) {
        cout << i << "\t" << epspLookup[i] << endl;
    }

    /*cout << "\n\nedPspLookup:\n\n";
    for (unsigned int i=0; i<pspLSize; ++i) {
        cout << i << "\t" << edPspLookup[i] << endl;
    }*/

    cout << "\n\nPrinting inhibitory tables:\nipspLookup:\n\n";
    for (unsigned int i=0; i<pspLSize; ++i) {
        cout << i << "\t" << ipspLookup[i] << endl;
    }

    /*cout << "\n\nidPspLookup:\n\n";
    for (unsigned int i=0; i<pspLSize; ++i) {
        cout << i << "\t" << idPspLookup[i] << endl;
    }*/
}

void AlphaNeuron::InitLookup()
{
	FunctionLookup* fl = Network::GetNetworkRef()->GetFunctionLookup();

	try {
		TableProperties t0 = GetTableProps(0);
		TableProperties t1 = GetTableProps(1);
		TableProperties t2 = GetTableProps(2);
		TableProperties t3 = GetTableProps(3);
		epspLookup = fl->GetTableData(t0);
		edPspLookup = fl->GetTableData(t1);
		ipspLookup = fl->GetTableData(t2);
		idPspLookup = fl->GetTableData(t3);
	}
	catch (TableNotFoundException& e) {
		//pspLookup = fl->MakeLookupTable(props);
		MakeLookupTables();
		//PrintLookupTables();
	}
}

TableProperties AlphaNeuron::GetTableProps(unsigned int index)
{
	TableProperties props;
	props.SetClassName("AlphaNeuron");
	props.SetTableSize(pspLSize);
	AlphaNeuronProperties* anProps = dynamic_cast<AlphaNeuronProperties*>(neuronProps);
	float membraneConst = anProps->GetMembraneConst();
	float eSynConst = anProps->GetExcitatorySynapseConst();
	float iSynConst = anProps->GetInhibitorySynapseConst();
	props.AddParam((AmTimeInt)index);
	props.AddParam(membraneConst);
	props.AddParam(eSynConst);
	props.AddParam(iSynConst);
	return props;
}

void AlphaNeuron::MakeLookupTables()
{
	FunctionLookup* fl = Network::GetNetworkRef()->GetFunctionLookup();
	
	TableProperties t0 = GetTableProps(0);
	TableProperties t1 = GetTableProps(1);
	TableProperties t2 = GetTableProps(2);
	TableProperties t3 = GetTableProps(3);
	LOGGER(6, "Making lookup tables")
    epspLookup = fl->MakeLookupTable(t0);
    edPspLookup = fl->MakeLookupTable(t1);
    ipspLookup = fl->MakeLookupTable(t2);
    idPspLookup = fl->MakeLookupTable(t3);
    LOGGER(6, "Lookups made")
    
	AlphaNeuronProperties* anProps = dynamic_cast<AlphaNeuronProperties*>(neuronProps);
	float membraneConst = anProps->GetMembraneConst();
	float eSynConst = anProps->GetExcitatorySynapseConst();
	float iSynConst = anProps->GetInhibitorySynapseConst();

	// TODO: Need to add in code for iSynConst here and in Euler
	Euler* eul = new Euler(1, 0, 100000, eSynConst, iSynConst, membraneConst);

	LOGGER(6, "Filling lookup tables")
    eul->GenExcitatoryTable(pspStepSize, epspLookup, edPspLookup);
    eul->GenInhibitoryTable(pspStepSize, ipspLookup, idPspLookup);
    
    delete eul;
    ScaleFunctionLookups();
}
