/***************************************************************************
                          basicneuron.cpp  -  description
                             -------------------
    copyright            : (C) 2000, 2001, 2002 by Matt Grover
    email                : mgrover@amygdala.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <cmath>
#include <iostream>
#include "basicneuron.h"
#include "functionlookup.h"
#include "network.h"
#include "dendrite.h"
#include "logging.h"
#include "utilities.h"
#include "physicalproperties.h"

using namespace std;
using namespace Amygdala;

unsigned int BasicNeuron::pspStepSize = 0;
unsigned int BasicNeuron::pspLSize = 0;

BasicNeuronProperties::BasicNeuronProperties():
    SpikingNeuronProperties(),
    synTimeConst(10.0)
{}

BasicNeuronProperties::BasicNeuronProperties(bool initializePhysicalProps):
    SpikingNeuronProperties(initializePhysicalProps),
    synTimeConst(10.0)
{
    if(initializePhysicalProps){
        PhysicalProperties::RGB color = {55, 200, 200};
        physProps->SetBodyColor(color);
    }
}

BasicNeuronProperties::BasicNeuronProperties(const BasicNeuronProperties& rhs):
    SpikingNeuronProperties(rhs)
{
    synTimeConst = rhs.synTimeConst;
}

BasicNeuronProperties::~BasicNeuronProperties()
{}

BasicNeuronProperties* BasicNeuronProperties::Copy() const
{
	return new BasicNeuronProperties(*this);
}

std::map< std::string, std::string > BasicNeuronProperties::GetPropertyMap() const
{
     std::map< std::string, std::string > props = SpikingNeuronProperties::GetPropertyMap();
     props["synTimeConst"] = Utilities::ftostr(synTimeConst);
     return props;
}

///////// End BasicNeuronProperties //////////


BasicNeuron::BasicNeuron(AmIdInt neuronId, const BasicNeuronProperties& neuronProps):
    SpikingNeuron(neuronId, neuronProps),
    histBeginIdx(0),
    maxThreshCrs(0),
    pspLookup(0),
    dPspLookup(0)
{
    if (!pspStepSize) {
	    Init();
    }
    
    InitLookup();
}

BasicNeuron::~BasicNeuron()
{
}

void BasicNeuron::SetProperties(BasicNeuronProperties* props)
{
	delete neuronProps;
	neuronProps = props->Copy();
}

void BasicNeuron::SpikeCleanup()
{
	// reset inputHist vector
	inputHist.clear();
	histBeginIdx = 0;
}

void BasicNeuron::CalcState(float& state, float& deriv,
                        const AmTimeInt& calcTime,
                        const AmTimeInt& historySize)
{
    AmTimeInt i, funcTime, tblIndex;
    float funcWeight, stepSizeInv = 1.0/pspStepSize;
    InputHist input;

    state = 0.0;
    deriv = 0.0;
    for (i=histBeginIdx; i<historySize; i++) {
        input = inputHist[i];
        funcTime = calcTime - input.time;
        funcWeight = input.weight;

        #ifdef DEBUG_NEURON
        cout << "funcTime: " << funcTime << endl;
        cout << "funcWeight: " << funcWeight << endl;
        #endif

        tblIndex = (unsigned int)(funcTime * stepSizeInv);
        if (tblIndex < pspLSize) {  // don't go beyond the end of the tables
            state = state + (funcWeight * (pspLookup[tblIndex]));
            deriv = deriv + (funcWeight * (dPspLookup[tblIndex]));
        }
        else {
            ++histBeginIdx;
        }
    }
}


void BasicNeuron::ProcessInput(const AmTimeInt& inTime)
{
    // If the neuron is within a refractory period,
    if ( (inTime - spikeTime) <= refPeriod ) {
        if (inTime > refPeriod) {
	        dendrite->ResetTrigger();
            return;
        }
    }

    unsigned int i, iterate, converged, histSize, tblIndex;
    AmTimeInt calcTime, funcTime;
    float currState = 0.0;
    float currDeriv = 0.0;
    float stateDelta = 0.0;
    float threshCrs = 0.0;
    float lstThreshCrs = 0.0;
    InputHist tmpInput;

    iterate = 1;
    converged = 0;
    calcTime = 0;
    funcTime = 0;
    histSize = 0;
    tblIndex = 0;

    if (!maxThreshCrs) {
        maxThreshCrs = pspLSize * pspStepSize;

        // find the convergence resolution (the resolution at which two values
        // of threshCrs are considered to be identical) -- must be <= simStepSize
        if (simStepSize > pspStepSize) {
            if (pspStepSize > (simStepSize / 2.0)) {
                convergeRes = pspStepSize;
            }
            else {
                convergeRes = simStepSize / 2.0;
            }
        }
        else {
            convergeRes = simStepSize;
        }
    }

    calcTime = inTime;
    inputTime = inTime;
    currTime = inTime;
    float inWeightPos, inWeightNeg;

    dendrite->GetStimulationLevel(inWeightPos, inWeightNeg);
    
    // Commented out until training classes are complete
    /*if (trainingMode) {
        dendrite->GetStimulationLevel(inWeightPos, inWeightNeg);
    }
    else {
        dendrite->GetStimulationLevel(inWeightPos, inWeightNeg);
    }*/


    // neg and pos weights are combined here because excitatory and
    // inhibitory synapses behave identically in this neuron model
    tmpInput.weight = inWeightPos + inWeightNeg;
    tmpInput.time = inTime;
    inputHist.push_back(tmpInput);

    histSize = inputHist.size();

    // Return if the neuron is going to spike during the current time
    // step and the input weight is positive
    if (inTime == schedSpikeTime) {
        if (tmpInput.weight > 0.0) {
            return;
        }
    }

    #ifdef DEBUG_NEURON
    cout << "\nNeuron " << nId << " receiving a spike.\n";
    for (i=0; i<histSize; i++) {
        tmpInput = inputHist[i];
        cout << "inTimeHist[" << i << "] " << tmpInput.time << endl;
        cout << "inWeightHist[" << i << "] " << tmpInput.weight << endl;
    }

    cout << "calcTime: " << calcTime << endl;
    cout << "inTime: " << inTime << endl;
    cout << "currTime: " << currTime << endl;
    #endif

    i = 0;

    // Use Newton's method to determine if and when the spike will occur
    /**************************************************************************
    *    1) Determine the membrane potential (currState) at time (calcTime -
    *       inTimeHist[i]) by summing the state of each inTimeHist[]
    *       (use pspLookup).
    *    2) Find the derivative of the function for calcTime (dPspLookup).
    *    3) Calculate intercept with thresholdPtnl.
    *    4) Set new calcTime to time of intercept.
    *    5) Repeat until:
    *        a) Two successive iterations result in a change in calcTime
    *           smaller than convergeRes (the convergence resolution).
    *            (Converges)
    *        b) The derivative of the function becomes negative.
    *            (Does not converge)
    **************************************************************************/
    // Take care of a special case first:
    // If this is the first input (starting from a rest potential)
    // then the PSP can only cross the threshold at t = synTimeConst
    // because of constraints on the maximum value of weights.
    #ifdef DEBUG_NEURON
    cout << "Starting main loop...\n";
    #endif

    if (histSize == 1) {
        iterate = 0;
        funcTime = (AmTimeInt)(dynamic_cast<BasicNeuronProperties*>(neuronProps)->GetSynapseConst())*1000;
        tblIndex = (funcTime / pspStepSize);
        currState = (inputHist[0].weight) * (pspLookup[tblIndex]);
        if (currState >= 1.0) {
            calcTime += funcTime;
            converged = 1;
        }
        else {
            converged = 0;
        }
    }
    else {
        // find a good starting point for N's method
//        calcTime = inTime + int( memTimeConst * 1000.0 );
//
//        while (iterate) {
//            CalcState(currState, currDeriv, calcTime, histSize);
//
//            if ( (currDeriv < 0.0) || (currState > thresholdPtnl) ) {
//                // went too far -- choose a smaller starting point
//                calcTime -= ( ( calcTime - inTime ) / 2 );
//            }
//            else {
//                // good starting point
//                iterate = 0;
//            }
//        }
//
//        iterate = 1;
    }

    lstThreshCrs = float(calcTime);
    while (iterate) {
        currState = 0.0;
        currDeriv = 0.0;
        // debug msg

        // calcTime has to be rounded to the nearest simStepSize
        Utilities::RoundTime(calcTime, pspStepSize);

        #ifdef DEBUG_NEURON
        cout << "calcTime: " << calcTime << endl;
        #endif

        // should be inline
        CalcState(currState, currDeriv, calcTime, histSize);
//        for (i=0; i<histSize; i++) {
//            funcTime = calcTime - inTimeHist[i];
//            funcWeight = inWeightHist[i];
//
//            // debugging msg
//            cout << "funcTime: " << funcTime << endl;
//            cout << "funcWeight: " << funcWeight << endl;
//            // end debugging
//
//            tblIndex = (funcTime / pspStepSize);
//            currState = currState + (funcWeight * (pspLookup[tblIndex]));
//            currDeriv = currDeriv + (funcWeight * (dPspLookup[tblIndex]));
//        }
        #ifdef DEBUG_NEURON
        cout << "currState: " << currState << endl;
        cout << "currDeriv: " << currDeriv << endl;
        #endif

        if ( (currDeriv < 0.0) && (currState < 1.0) ) {
            converged = 0;
            iterate = 0;

            // set the scheduled spike time to zero to keep
            // the neuron from spiking when an inhibitory input
            // spike has canceled a previously scheduled spike.
            schedSpikeTime = 0;
        }
        else if (currState > 1.0) {
            converged = 1;
            iterate = 0;
        }
        else {
            stateDelta = fabs(1.0 - currState);
            threshCrs = (stateDelta / currDeriv) + lstThreshCrs;

            #ifdef DEBUG_NEURON
            cout << "stateDelta: " << stateDelta << endl;
            cout << "threshCrs: " << threshCrs << endl;
            cout << "lstThreshCrs: " << lstThreshCrs << endl;
            #endif

            if ( (threshCrs - lstThreshCrs) < convergeRes) {
                converged = 1;
                iterate = 0;

                #ifdef DEBUG_NEURON
                cout << "threshCrs - lstThreshCrs < convergeRes\n";
                cout << "Converged\n";
                #endif
            }
            else {
                converged = 0;
                if (threshCrs > (maxThreshCrs + lstThreshCrs)) {
                    iterate = 0;
                }
                calcTime = int(threshCrs);
                lstThreshCrs = threshCrs;

                // set the scheduled spike time to zero to keep
                // the neuron from spiking when an inhibitory input
                // spike has canceled a previously scheduled spike.
                schedSpikeTime = 0;
            }

        }
    }
    if (converged) {
        // Make sure we wait at least one cycle before spiking
        if (calcTime == inTime) {
            calcTime += simStepSize;
        }
        schedSpikeTime = calcTime;

        LOGGER(5, "Neuron " << nId << " Scheduling spike at " << schedSpikeTime)
        
		Network::GetNetworkRef()->ScheduleSpike( schedSpikeTime, this );
    }
}

void BasicNeuron::ScaleFunctionLookups()
{
    unsigned int i;
    float lastPtnl=0.0;
    float maxPtnl=0.0;

    cout << "Scaling function lookup\n";
    for (i=0; i<pspLSize; i++) {
        LOGGER(1, "pspLookup[" << i << "] = " << pspLookup[i]);
    }
    for (i=0; i<pspLSize; ++i) {
        maxPtnl = pspLookup[i];

        if (maxPtnl < lastPtnl) {
            maxPtnl = lastPtnl;
            break;
        }
        else {
            lastPtnl = maxPtnl;
        }
    }

    if (maxPtnl == 0.0)
        return;

    // Normalize the function
    float scaleFactor = 1.0/maxPtnl;
    for (i=0; i<pspLSize; i++) {
        pspLookup[i] = scaleFactor*pspLookup[i];
        dPspLookup[i] = scaleFactor*dPspLookup[i];
    }
}

float BasicNeuron::GetMembranePtnl()
{
    float currState=0.0, currDeriv=0.0;
    AmTimeInt calcTime = Network::GetNetworkRef()->SimTime();
    Utilities::RoundTime(calcTime, pspStepSize);
    CalcState(currState, currDeriv, calcTime, inputHist.size());
    return currState;
}

void BasicNeuron::Init()
{
	pspStepSize = Network::TimeStepSize();
	pspLSize = 100000/pspStepSize;		// 100ms / step size
	
	LOGGER(3, "pspStepSize: " << pspStepSize << " pspLSize: " << pspLSize)
}

void BasicNeuron::InitLookup()
{
	FunctionLookup* fl = Network::GetNetworkRef()->GetFunctionLookup();

	try {
		TableProperties t0 = GetTableProps(0);
		TableProperties t1 = GetTableProps(1);
		pspLookup = fl->GetTableData(t0);
		dPspLookup = fl->GetTableData(t1);
	}
	catch (TableNotFoundException& e) {
		MakeLookupTables();
		//PrintLookupTables();
	}
}

TableProperties BasicNeuron::GetTableProps(unsigned int index)
{
	TableProperties props;
	props.SetClassName("BasicNeuron");
	props.SetTableSize(pspLSize);
	BasicNeuronProperties* bnProps = dynamic_cast<BasicNeuronProperties*>(neuronProps);
	float synConst = bnProps->GetSynapseConst();
	props.AddParam((AmTimeInt)index);
	props.AddParam(synConst);
	return props;
}

void BasicNeuron::MakeLookupTables()
{
	FunctionLookup* fl = Network::GetNetworkRef()->GetFunctionLookup();
	
	TableProperties t0 = GetTableProps(0);
	TableProperties t1 = GetTableProps(1);
	LOGGER(6, "Making lookup tables")
    pspLookup = fl->MakeLookupTable(t0);
    dPspLookup = fl->MakeLookupTable(t1);
    LOGGER(6, "Lookups made")
    
	LOGGER(6, "Filling lookup tables")
    float calcTm = 0.0;
    for (unsigned int i=0; i<pspLSize; ++i) {
        calcTm = (float(i) * pspStepSize);
        PspKernel(calcTm, &pspLookup[i]);
    }
    for (unsigned int i=0; i<pspLSize; ++i) {
        calcTm = (float(i) * pspStepSize);
        DPspKernel(calcTm, &dPspLookup[i]);
    }
	
    ScaleFunctionLookups();
}

void BasicNeuron::PspKernel(float calcTime, float* pspElement)
{
    // Model the PSP as Ep = ( t/( Ts^2 ) ) * exp[ -t / Ts ]
    float term1 = 0.0, term2 = 0.0, expTerm = 0.0, convCalcTime;

	BasicNeuronProperties* bnProps = dynamic_cast<BasicNeuronProperties*>(neuronProps);
	float synConst = bnProps->GetSynapseConst();

    // convert time from microseconds to milliseconds
    convCalcTime = calcTime*0.001;

    term1 = (convCalcTime/(pow(synConst, 2)));
    expTerm = -convCalcTime/synConst;
    term2 = exp(expTerm);
    *pspElement = term1 * term2;
}

void BasicNeuron::DPspKernel(float calcTime, float* dPspElement)
{
    float tempVar, convCalcTime;

    // convert time from microseconds to milliseconds
    convCalcTime = calcTime*0.001;

	BasicNeuronProperties* bnProps = dynamic_cast<BasicNeuronProperties*>(neuronProps);
	float synConst = bnProps->GetSynapseConst();

    tempVar = exp(-convCalcTime/synConst)/(pow(synConst, 2));
    tempVar = tempVar - (convCalcTime/(pow(synConst, 3))*exp(-convCalcTime/synConst));

    // Divide the result by 1000 so that it has units of mV/us (converted from
    // mV/ms).
    *dPspElement = tempVar*0.001;
}

