/***************************************************************************
                          lsmtopology.cpp  -  description
                             -------------------
    begin                : Wed Mar 30 2005
    copyright            : (C) 2005 by Rudiger Koch
    email                : rkoch@amygdala.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "lsmtopology.h"
#include "spikeinput.h"
#include "factory.h"
#include "physicalproperties.h"
#include "neuronproperties.h"
#include "neuron.h"
#include "spikingneuron.h"
#include "inputneuron.h"
#include "nconnector.h"
#include "synapse.h"
#include "synapseproperties.h"
#include "utilities.h"
#include "logging.h"

#include <stdexcept>
#include <cstdlib>

namespace Amygdala {

using namespace std;
using namespace Utilities;

LSMTopology::LSMTopology(string name) : 
    Topology(name),
    X(0), Y(0), Z(0),
    lastInX(0), lastInY(0), lastInZ(0), lastInSide(0),
    lastOutX(0), lastOutY(0), lastOutZ(0)
{
}


LSMTopology::~LSMTopology()
{}


Topology * LSMTopology::MakeInput(SpikeInput * sInput, unsigned int iNeurons)
{
    if( X==0 || Y==0 || Z==0) throw runtime_error("No cube defined yet. Call MakeCube before attempting to create Input");
    //TFactory * lsmInput = dynamic_cast<TFactory*> (Factory::GetRegistry().GetFactory("LSMTopology::Input"));
    //if(!lsmInput) throw runtime_error("No LSMTopology::Input factory registered");
    TopologyFactory<LSMTopology::Input> lsmInput;
    Input* in = lsmInput("LSM Input");
    input.push_back(in);
    
    // Move the code that used to be in Input::MakeInput() to here since more than one Input
    // object is now allowed.  Since we have to keep track of where input neurons have been added
    // previously, it will be easier to do that here.
    //in->MakeInput(sInput, X, Y, Z, iNeurons);
    LOGGER(3, "Making Inputs");
    unsigned int neuronCounter = 1;
    unsigned int inputLocs = 2*X*Z + 2*Y*Z;
    unsigned int totInputs = 0;
    
    for (unsigned int i=0; i<input.size(); ++i) {
        totInputs += input[i]->size();
    }
    totInputs += iNeurons;
    if(totInputs > inputLocs) 
            throw runtime_error("Insufficient space for InputNeurons. Consider making the cube larger");
    
    if (lastInSide == 0) {
        int y=-1;
        // distribute the InputNeurons alongside at the X-Z side of the cube (y == -1)
        for(int x=lastInX; x < X; x++){
            for(int z=lastInZ; z < Z; z++){
                InputNeuron * inNrn = in->MakeInputNeuron(x, y, z);
                sInput->AddNeuron(inNrn);
                lastInZ = z;
                if(++neuronCounter > iNeurons) return in;
            }
            lastInZ = 0;   
            lastInX = x;
        }
        lastInX = 0;
        lastInZ = 0;
        lastInSide = 1;
    }
    
    if (lastInSide == 1) {
        int y=Y;
        // distribute the InputNeurons alongside at the X-Z side of the cube (y == Y)
        for(int x=lastInX; x < X; x++){
            for(int z=lastInZ; z < Z; z++){
                InputNeuron * inNrn = in->MakeInputNeuron(x, y, z);
                sInput->AddNeuron(inNrn);
                lastInZ = z;
                if(++neuronCounter > iNeurons) return in;
            }   
            lastInZ = 0;
            lastInX = x;
        }
        lastInX = 0;
        lastInZ = 0;
        lastInSide = 2;
    }
    
    if (lastInSide == 2) {
        // if there are still more input neurons
        // distribute them alongside at the X-Y side of the cube (z == -1)
        int z=-1;
        for(int x=lastInX; x < X; x++){
            for(int y=lastInY; y < Y; y++){
                InputNeuron * inNrn = in->MakeInputNeuron(x, y, z);
                sInput->AddNeuron(inNrn);
                lastInY = y;
                if(++neuronCounter > iNeurons) return in;
            }   
            lastInY = 0;
            lastInX = x;
        }
        lastInX = 0;
        lastInSide = 3;
    }
    
    if (lastInSide == 3) {
        // if there are still more input neurons
        // distribute them alongside at the X-Y side of the cube (z == Z)
        int z=Z;
        for(int x=lastInX; x < X; x++){
            for(int y=lastInY; y < Y; y++){
                InputNeuron * inNrn = in->MakeInputNeuron(x, y, z);
                sInput->AddNeuron(inNrn);
                lastInY = y;
                if(++neuronCounter > iNeurons) return in;
            }   
            lastInY = 0;
            lastInX = x;
        }
        lastInX = 0;
        lastInSide = 4;
    }
    
    return in;
}

Topology * LSMTopology::MakeReadout(const string & nType, unsigned int oNeurons)
{
    if( X==0 || Y==0 || Z==0) throw runtime_error("No cube defined yet. Call MakeCube before attempting to create Output");
    TFactory * lsmOutput = dynamic_cast<TFactory*> (Factory::GetRegistry().GetFactory("LSMTopology::Output"));
    if(!lsmOutput) throw runtime_error("No LSMTopology::Output factory registered");
    Output* out = dynamic_cast<Output*> (lsmOutput->MakeTopology("LSM Output"));
    output.push_back(out);
    
    // Move the code that used to be in Output::MakeReadout() to here since more than one Output
    // object is now allowed.  Since we have to keep track of where neurons have been added
    // previously, it will be easier to do that here.
    //out->MakeReadout(nType, X, Y, oNeurons);
    LOGGER(3, "Making Outputs");
    unsigned int neuronCounter = 1;

    NFactory * nf =  dynamic_cast<NFactory*> (Factory::GetRegistry().GetFactory(nType));
    if (nf == NULL) throw runtime_error("No neuron type " + nType + " registered!");

    NeuronProperties * nProps = nf->MakeNeuronProperties(true);
    PhysicalProperties * pProps = nProps->GetPhysicalProps();

    // distribute the OutputNeurons alongside at the Y-Z side of the cube starting at X+1, 0, 0
    for (int z=lastOutZ; true; z++){
        for (int y = lastOutY; y < Y; y++){
            pProps->SetPosition(X + 1, y, z);
            Neuron * n = nf->MakeNeuron(Network::GetNetworkRef()->GetNewNeuronId(), nProps, out);
            OutputManager::AddNeuronToGroup(n, OUTPUTGROUP);
            LOGGER(4, "Output neuron at " << X+1 << ", " << y << ", " << z);
            lastOutY++;
            if(++neuronCounter > oNeurons) return out;
        }
        lastOutY=0;
        lastOutZ++;
    }
    
    return out;
}

Topology * LSMTopology::MakeReadout(NFactory* nf, NeuronProperties* np, unsigned int oNeurons)
{
    if( X==0 || Y==0 || Z==0) throw runtime_error("No cube defined yet. Call MakeCube before attempting to create Output");
    TFactory * lsmOutput = dynamic_cast<TFactory*> (Factory::GetRegistry().GetFactory("LSMTopology::Output"));
    if(!lsmOutput) throw runtime_error("No LSMTopology::Output factory registered");
    Output* out = dynamic_cast<Output*> (lsmOutput->MakeTopology("LSM Output"));
    output.push_back(out);
    
    // Move the code that used to be in Output::MakeReadout() to here since more than one Output
    // object is now allowed.  Since we have to keep track of where neurons have been added
    // previously, it will be easier to do that here.
    //out->MakeReadout(nType, X, Y, oNeurons);
    LOGGER(3, "Making Outputs");
    unsigned int neuronCounter = 1;

    /*NFactory * nf =  dynamic_cast<NFactory*> (Factory::GetRegistry().GetFactory(nType));
    if (nf == NULL) throw runtime_error("No neuron type " + nType + " registered!");*/

    //NeuronProperties * nProps = nf->MakeNeuronProperties(true);
    PhysicalProperties * pProps = np->GetPhysicalProps();

    // distribute the OutputNeurons alongside at the Y-Z side of the cube starting at X+1, 0, 0
    for (int z=lastOutZ; true; z++){
        for (int y = lastOutY; y < Y; y++){
            pProps->SetPosition(X + 1, y, z);
            Neuron * n = nf->MakeNeuron(Network::GetNetworkRef()->GetNewNeuronId(), np, out);
            OutputManager::AddNeuronToGroup(n, OUTPUTGROUP);
            LOGGER(4, "Output neuron at " << X+1 << ", " << y << ", " << z);
            lastOutY++;
            if(++neuronCounter > oNeurons) return out;
        }
        lastOutY=0;
        lastOutZ++;
    }
    
    return out;
}

void LSMTopology::MakeCube(const string & nType, unsigned int lx, unsigned int ly, unsigned int lz)
{
    X=lx; Y=ly; Z=lz;
    LOGGER(1, "Making Cube of size " << X << ", " << Y << ", " << Z);
    
    NFactory * nf =  dynamic_cast<NFactory*> (Factory::GetRegistry().GetFactory(nType));
    if (nf == NULL) throw runtime_error("No neuron type " + nType + " registered!");
    
    NeuronProperties * nProps = nf->MakeNeuronProperties(true);
    PhysicalProperties * pProps = nProps->GetPhysicalProps();
    
    for (int x = 0; x < X; x++){
        for (int y = 0; y < Y; y++){
            for (int z = 0; z < Z; z++){
                pProps->SetPosition(x, y, z);
                Neuron * n = nf->MakeNeuron(Network::GetNetworkRef()->GetNewNeuronId(), nProps, this);
                OutputManager::AddNeuronToGroup(n, HIDDENGROUP);
                if(RandUnit() <= 0.2 ){
                    n->Inhibitory(true);
                    LOGGER(4, "Inhibitory neuron at " << x << ", " << y << ", " << z);
                } else {
                    n->Inhibitory(false);
                    LOGGER(4, "Excitatory neuron at " << x << ", " << y << ", " << z);
                }
            }
        }
    }
}

void LSMTopology::MakeCube(NFactory* nf, NeuronProperties* np, unsigned int lx, unsigned int ly, unsigned int lz)
{
    X=lx; Y=ly; Z=lz;
    LOGGER(1, "Making Cube of size " << X << ", " << Y << ", " << Z);
    
    //NeuronProperties * nProps = nf->MakeNeuronProperties(true);
    PhysicalProperties * pProps = np->GetPhysicalProps();
    
    for (int x = 0; x < X; x++){
        for (int y = 0; y < Y; y++){
            for (int z = 0; z < Z; z++){
                pProps->SetPosition(x, y, z);
                Neuron * n = nf->MakeNeuron(Network::GetNetworkRef()->GetNewNeuronId(), np, this);
                OutputManager::AddNeuronToGroup(n, HIDDENGROUP);
                if(RandUnit() <= 0.2 ){
                    n->Inhibitory(true);
                    LOGGER(4, "Inhibitory neuron at " << x << ", " << y << ", " << z);
                } else {
                    n->Inhibitory(false);
                    LOGGER(4, "Excitatory neuron at " << x << ", " << y << ", " << z);
                }
            }
        }
    }
}


void LSMTopology::Connect(const float lambda, 
                          const EI & A, 
                          const EI & delay,
                          const EI & C,
                          const EI & U,
                          const EI & D,
                          const EI & F,
                          const  string & sType,
						  const float likelyhood)
{
    LOGGER(1, "Connecting the liquid with synapse type " << sType);
    int cType;
    ConnectorRegistry & cr = ConnectorRegistry::GetRegistry();
    NConnector * connector = cr.GetConnector(sType);
    if(!connector) throw runtime_error("No connector for synapse type " + sType);
    for (iterator itPre=begin(); itPre != end(); itPre++){
       for (iterator itPost=begin(); itPost != end(); itPost++){
           Neuron * pre = *itPre;
           SpikingNeuron * post = dynamic_cast <SpikingNeuron*> (*itPost);
           if(!post) continue; // Postneuron is an InputNeuron
           if(pre == post) continue;  // no self - connection
           
           if(!pre->Inhibitory() && !post->Inhibitory()) cType = 0;
           else if(!pre->Inhibitory() && post->Inhibitory()) cType = 1;
           else if(pre->Inhibitory() && !post->Inhibitory()) cType = 2;
           else  cType = 3;
           
           PhysicalProperties * pPre = pre->Properties()->GetPhysicalProps();
           PhysicalProperties * pPost = post->Properties()->GetPhysicalProps();
           float dd = (pPre->GetX() - pPost->GetX()) * (pPre->GetX() - pPost->GetX()) +
                      (pPre->GetY() - pPost->GetY()) * (pPre->GetY() - pPost->GetY()) +
                      (pPre->GetZ() - pPost->GetZ()) * (pPre->GetZ() - pPost->GetZ());
           float cChance = C.EIV[cType] * exp(-dd/(lambda*lambda));
           if(RandUnit() < cChance ){  // connect
               LOGGER(4, "Connecting " << pre->GetId() << " --> " << post-> GetId());
               DynamicSynapseProperties * sProps = dynamic_cast < DynamicSynapseProperties * > 
                                                                 (connector->GetDefaultProperties());
               if(!sProps){
                   cerr << __FILE__ << ":" << __LINE__ << ": Cannot convert SynapseProperties" << endl;
                   continue;
               }
               sProps->SetDelay(AmTimeInt(delay.EIV[cType]));
            // TODO: The weight is gamma distributed
               float tmp=-1;
               /*while (tmp<0) {
                   tmp = Utilities::GaussRand(A.EIV[cType], A.EIV[cType]);
               }*/
               sProps->SetDynaParam('A', A.EIV[cType]);
            // TODO: The following values are Gauss distributed
               /*tmp=-1;
               while (tmp<0) {
                   tmp = Utilities::GaussRand(U.EIV[cType], U.EIV[cType]/2);
               }*/
               sProps->SetDynaParam('U', U.EIV[cType]);
               /*tmp=-1;
               while (tmp<0) {
                   tmp = Utilities::GaussRand(D.EIV[cType], D.EIV[cType]/2);
               }*/
               sProps->SetDynaParam('D', D.EIV[cType]);
               /*tmp=-1;
               while (tmp<0) {
                   tmp = Utilities::GaussRand(F.EIV[cType], F.EIV[cType]/2);
               }*/
               sProps->SetDynaParam('F', F.EIV[cType]);
               connector->Connect(pre, post, *sProps);
           }
       }
    }
    
    if(output.size()){
        for (unsigned int i=0; i<output.size(); ++i) {
            NConnector * connectorOut = ConnectorRegistry::GetRegistry().GetConnector("StaticSynapse");
            if(!connectorOut) throw runtime_error("No connector for synapse type StaticSynapse" );
            for (iterator itPre=begin(); itPre != end(); itPre++){
                for (iterator itPost=output[i]->begin(); itPost != output[i]->end(); itPost++){
                    SpikingNeuron * pre = dynamic_cast <SpikingNeuron*> (*itPre);
                    if(!pre) throw runtime_error(string("Neuron: ") +  Utilities::itostr(pre->GetId()) +
                                                        ": Not a SpikingNeuron"); 
                    SpikingNeuron * post = dynamic_cast <SpikingNeuron*> (*itPost);
                    if(!post) throw runtime_error(string("Neuron: ") +  Utilities::itostr(post->GetId()) +
                                                        ": Not a SpikingNeuron"); 
                    StaticSynapseProperties * sProps = dynamic_cast < StaticSynapseProperties * > 
                                                                    (connectorOut->GetDefaultProperties());
                    float w = (RandUnit())*0.2;
                    if (RandUnit()<0.5) {
                        w = w * -1.;    // make half the connections inhibitory
                    }
                    sProps->SetWeight(w);
                    connectorOut->Connect(pre, post, *sProps);                
                }
            }
        }
    } else {
        LOGGER(1, "No output defined");
    }

    if(input.size()){
        for (unsigned int i=0; i<input.size(); ++i) {
            NConnector * connectorIn = ConnectorRegistry::GetRegistry().GetConnector("StaticSynapse");
            if(!connectorIn) throw runtime_error("No connector for synapse type StaticSynapse" );
            for (iterator itPre=input[i]->begin(); itPre != input[i]->end(); itPre++){
                for (iterator itPost = begin(); itPost != end(); itPost++){
                    if(RandUnit() > likelyhood) continue;  // draw random neurons from the liquid
                    SpikingNeuron * post = dynamic_cast <SpikingNeuron*> (*itPost);
                    if(!post) throw runtime_error(string("Neuron: ") +  Utilities::itostr(post->GetId()) + ": Not a SpikingNeuron");
                    Neuron * pre = dynamic_cast <Neuron*> (*itPre);
                    StaticSynapseProperties * sProps = dynamic_cast < StaticSynapseProperties * > 
                                                                    (connectorIn->GetDefaultProperties());
                    if (post->Inhibitory()) {
                        sProps->SetWeight(0.6);
                    }
                    else {
                        sProps->SetWeight(1.2);
                    }
                    connectorIn->Connect(pre, post, *sProps);                
                }
            }
            /*for (iterator itPost = begin(); itPost != end(); itPost++){
                if(RandUnit() > likelyhood) continue;  // draw random neurons from the liquid
                SpikingNeuron * post = dynamic_cast <SpikingNeuron*> (*itPost);
                if(!post) throw runtime_error(string("Neuron: ") +  Utilities::itostr(post->GetId()) + ": Not a SpikingNeuron");
                for (iterator itPre=input[i]->begin(); itPre != input[i]->end(); itPre++){
                    Neuron * pre = dynamic_cast <Neuron*> (*itPre);
                    StaticSynapseProperties * sProps = dynamic_cast < StaticSynapseProperties * > 
                                                                    (connectorIn->GetDefaultProperties());
                    sProps->SetWeight(1.2);
                    connectorIn->Connect(pre, post, *sProps);                
                }
            }*/
        }
    } else {
        LOGGER(1, "No input defined");
    }
}



// Local class Output

LSMTopology::Output::Output(string name) : Topology(name)
{}
LSMTopology::Output::~Output()
{}

/*void LSMTopology::Output::MakeReadout(const string & nType, unsigned int X, unsigned int Y, unsigned int oNeurons)
{
    LOGGER(3, "Making Outputs");
    unsigned int neuronCounter = 1;

    NFactory * nf =  dynamic_cast<NFactory*> (Factory::GetRegistry().GetFactory(nType));
    if (nf == NULL) throw runtime_error("No neuron type " + nType + " registered!");

    NeuronProperties * nProps = nf->MakeNeuronProperties(true);
    PhysicalProperties * pProps = nProps->GetPhysicalProps();

    // distribute the OutputNeurons alongside at the Y-Z side of the cube starting at X+1, 0, 0
    for (int z=0; true; z++){
        for (int y = 0; y < Y; y++){
            pProps->SetPosition(X + 1, y, z);
            Neuron * n = nf->MakeNeuron(Network::GetNetworkRef()->GetNewNeuronId(), nProps, this);
            OutputManager::AddNeuronToGroup(n, OUTPUTGROUP);
            LOGGER(4, "Output neuron at " << X+1 << ", " << y << ", " << z);
            if(++neuronCounter > oNeurons) return;
        }
    }
}*/


// Local class Input

LSMTopology::Input::Input(string name) : Topology(name)
{}
LSMTopology::Input::~Input()
{}

InputNeuron * LSMTopology::Input::MakeInputNeuron(int x, int y, int z)
{
    LOGGER(5, "Making Input Neuron at " << x << ", " << y << ", " << z);
    NFactory * iFactory = dynamic_cast<NFactory*>
                        (Factory::GetRegistry().GetFactory("InputNeuron"));
    if(iFactory == NULL) throw runtime_error("Cannot get a NeuronFactory for input");
    NeuronProperties * iProps = iFactory->MakeNeuronProperties(true);
    iProps->GetPhysicalProps()->SetPosition(x, y, z);
    AmIdInt id = Network::GetNetworkRef()->GetNewNeuronId();
    InputNeuron * n = dynamic_cast <InputNeuron *> (iFactory->MakeNeuron(id, iProps, this));
	OutputManager::AddNeuronToGroup(n, INPUTGROUP);
    if(n == NULL) throw runtime_error(string("Creating InputNeuron id ") + Utilities::itostr(id) + "failed");
    return n;
}

   

}; // namespace Amygdala

