/***************************************************************************
                          mouse.h  -  description
                             -------------------
    begin                : Thu Apr 4 2002
    copyright            : (C) 2002 by Rudiger Koch, Matt Grover
    email                : rkoch@rkoch.org, matt@amygdala.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef MOUSE_H
#define MOUSE_H


#include <vector>
#include <string>
#include <map>

#include <amygdala/network.h>
#include <amygdala/alphaneuron.h>
#include <amygdala/statisticsoutput.h>
#include <amygdala/flatlayer.h>
#include <amygdala/inputneuron.h>
#include <amygdala/factory.h>
#include "musspikeinput.h"

using namespace std;
using namespace Amygdala;

/**
  * Mus Silicium example. We assume that sound files have 12000 samples per second,
  * 16 bit per sample and 1 channel - this is the format of the wav files downloadable
  * from the <A HREF="http://moment.princeton.edu/~mus/Organism/">Mus Silicium homepage</A>
  *@author Rdiger Koch <rkoch@rkoch.org>
  *@author Matt Grover <matt@amygdala.org>
*/

class MouseOutput: public StatisticsOutput {
public:
    typedef SpikeOutputFactory<MouseOutput> Factory;
    MouseOutput():StatisticsOutput() {}
    virtual ~MouseOutput() {}

	/**
	 * Raise an output event.  If output is turned on in streaming mode,
	 * this function will be called every time an output neuron spikes.
	 */
	virtual void OutputEvent(Neuron* nrn, AmTimeInt eventTime);

	/**
	 * If output is turned on in caching mode,
	 * this function will be called every time the cache is emptied.
	 * Arguments TBD.
	 */
	virtual void OutputCache() {};
};

class Mouse {
public:

    enum BandType {BTYPE_ON, BTYPE_OFF, BTYPE_PEAK};

    struct Band {
        /** the type of event that neurons in this band respond to */
        BandType type;
        /** the band number */
        AmIdInt ID;
        /** The threashold of getting active */
        float threshold;
        /** Has this band been activated already? */
        bool active;
    };

    Mouse(float _maxSpikeRate=100.0,
        float _areaAWeight=1.0,
        float _areaWWeight=0.001,
        float _synapticConst=4.0,
        float _membraneConst=15.0);
    ~Mouse();
    /** Load a soundfile and convert it into a power spectrum */
//    void Load(const char *filename);
    /** Load an event file provided by the mus silicium web site */
    void LoadEvents(string filename);

    void LoadEvents(float* eventData);
    
    /** Run the simulation */
    void Run();
    /** Run the mouse with the loaded sound and training turned on. */
//    void Train(int iteration=0);

    /** Set up the network connections for the training pattern */
    void Train2();

protected: // Protected attributes
    vector <Band> fbands;   // frequency bands
    vector <Amygdala::FlatLayer*> areaA;  // areaA layers -- one layer per band (parallel layers)
    vector <float> decayTime;  // the decay time for the ith neuron in an area A band
    vector <float> eventTime;
    MusSpikeInput *spikeInput;
    MouseOutput* spikeOut;
    AmTimeInt maxRunTime;
    Amygdala::FlatLayer* alpha;   // excitatory neurons in area W
    Amygdala::FlatLayer* beta;    // inhibitory neurons in area W
    Amygdala::FlatLayer* gamma;   // output neurons

    const unsigned int neuronsPerBand;
    const unsigned int bands;
    const unsigned int lastLayerId;
    unsigned int layerCount;
    float maxSpikeRate;     // maximum spike rate in us
    float areaAWeight;
    float areaWWeight;
    float esynapticConst;
    float isynapticConst;
    float membraneConst;

//    vector< vector<float> > fullSpectrum;   // used for plotting

protected: // Protected methods
    /** Check if the event of a neuron is above the threshold */
//    int GetCurveShape(int index);

    /** Returns the result of the first derivative of Energy(t) */
//    bool CheckActivate(AmIdInt bId, float energy, int shape);

    /** Find the active input neurons for the current training pattern */
    vector<InputNeuron*> FindActiveNeurons();

    void InitOutput();

    /** Generate input from the event list */
    void ScheduleInput();

    /** Build the input Area */
    void BuildAreaA();

    /** Build area W (layers alpha, beta, and gamma */
    void BuildAreaW();

    /** Form connections between and within alpha and beta layers */
    void ConnectAlphaBeta();

    /** Return a random number between 0 and 100 */
    float RandPercent();
};

namespace Amygdala {
    namespace Factory {
        static MouseOutput::Factory MakeMouseOutput;
    }
}
#endif
