/***************************************************************************
                          genome.cpp  -  description
                             -------------------
    begin                : Sun Nov 25 2001
    copyright            : (C) 2001 by Rudiger Koch
    email                : rkoch@rkoch.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "genome.h"
#include "gfile.h"
#include "chromosome.h"
#include "amygdala/logging.h"

#include "config.h"

#include <stdexcept>
#include <fstream>
#include <sstream>
#include <iostream>
#include <string>
#include <stdio.h>

using namespace std;

map <string, string> Genome::config;

Genome::Genome(){
    LOGGER(3, "Making new genome");
}


Genome::~Genome(){
    LOGGER(3, "Deleting genome");
    for(unsigned int i=0; i<chromosomes.size(); i++) delete chromosomes[i];
}


/** Print out the current configuration */
void Genome::PrintConfig(){
    for(map < string, string>::iterator cfgIt = config.begin(); cfgIt != config.end(); cfgIt++){
        cout << cfgIt->first << " = " << cfgIt->second << endl;
    }
}

/** Load the config file */
void Genome::LoadConfig(){
    string line;

    string configfile = string(GENESERVER_DIR) + "/geneserver.conf";
        
    ifstream cfg(configfile.c_str(), ios::in);
    if(!cfg.good()) throw runtime_error(configfile + ": Cannot read");
    
    LOGGER(2, "Reading config file " << configfile);

    // defaults
    config["geneserver_dir"]              = GENESERVER_DIR;
    config["http_port"]                   = "10080";
    config["population"]                  = "30";
    config["survivors"]                   = "10";
    config["save_best"]                   = "0";
    config["likely_crossover"]            = "50";
    config["likely_mutation_flip"]        = "1";
    config["likely_mutation_pm"]          = "0";
    config["amount_mutation_pm"]          = "0";
    config["likely_mutation_duplication"] = "0";
    config["likely_mutation_deletion"]    = "0";
    config["mutation_heat"]               = "off";
    config["alltime_best"]                = "1000000000";

    while (!cfg.eof()){
        char c_line[200];
        cfg.getline(c_line, 199);
        line = string(c_line);
        if(line.length() == 0) continue;
        if(line[0] == '#') continue;
        string::size_type sep = line.find('=');
        if(sep == string::npos){
            cerr << "Bad config: " << configfile << endl;
            exit(-1);
        }
        string key = line.substr(0, sep);
        if(config.find(key) == config.end()) throw  runtime_error(key + ": No such key");
        config[line.substr(0, sep)] = line.substr(sep+1);
    }
}


string Genome::GetAttribute(string attr){
    if(config.find(attr) == config.end()) throw  runtime_error(attr + ": No such key");
    string value = config[attr];
    return value;
}


void Genome::Save(){
    GFile f(id);
    string genomeString;

    for(unsigned int i=0; i<chromosomes.size(); i++){
        genomeString += chromosomes[i]->ToString();
    }
    f.Save(FormatToString() + "\r\n" + genomeString);
}


/** Load a Genome from a file. The genome is stored in this object */
void Genome::Load(int _id){
    id = _id;
    GFile f(id);
    string fileString = f.Load();
    string::size_type crlf = fileString.find("\r\n");
    if(crlf == string::npos) throw runtime_error("Genome file has bad format");
    string formatString = fileString.substr(0, crlf);
    string genomeString = fileString.substr(crlf+2);
    chromosomes.clear();
    vector <ChromosomeFormat> chromosomeFormat = StringToFormat(formatString);

    int pos = 0;
    for(unsigned int i=0; i<chromosomeFormat.size(); i++) {
        int offset = chromosomeFormat[i].genes * (chromosomeFormat[i].gene_size + MUTATION_HEAT);
        string chromosomeString = genomeString.substr(pos, offset);
        pos += offset;
        Chromosome *c = new Chromosome(chromosomeFormat[i].genes, chromosomeFormat[i].gene_size,
                                       chromosomeFormat[i].mutationMode, chromosomeString);
        chromosomes.push_back(c);
    }
}



const int& Genome::GetId(){
    return id;
}


Genome* Genome::Mate(Genome *male){
    Genome* offspring = new Genome(this);
    for(unsigned int i=0; i<chromosomes.size(); i++){
        LOGGER(5, "Crossover");
        offspring->chromosomes[i]->Crossover(male->chromosomes[i]);
        LOGGER(5, "ID: " << offspring->GetId() << "Mutating chromosome " << i);
        offspring->chromosomes[i]->Mutate();
    }
    return offspring;
}


Genome::Genome(Genome* orig){
    LOGGER(3, "Copying genome" << orig->GetId());
    id = GFile::IncrementCounter();
    vector <ChromosomeFormat> chromosomeFormat = StringToFormat(orig->FormatToString());
    chromosomes.clear();
    for(unsigned int i=0; i<chromosomeFormat.size(); i++) {
        string cString = orig->chromosomes[i]->ToString();
        Chromosome *c = new Chromosome(chromosomeFormat[i].genes, chromosomeFormat[i].gene_size,
                                       chromosomeFormat[i].mutationMode, cString);
        chromosomes.push_back(c);
    }
}


string Genome::FormatToString(){
    ostringstream fmt;
    for(unsigned int i=0; i<chromosomes.size(); i++){
        fmt << '{' << chromosomes[i]->GetNumGenes() << ',' << 
                      chromosomes[i]->GetGeneSize() << ',' <<
                      chromosomes[i]->GetMutationMode() << '}';
    }
    return fmt.str();
}


void Genome::Die(){
    GFile f(id);
    f.Die();
}

void Genome::MateAll(int *fitest)
{
    int population = atoi(Genome::GetAttribute("population").c_str());
    int saveBest   = atoi(Genome::GetAttribute("save_best").c_str());
    int childsPerPair = 2 * population / atoi(Genome::GetAttribute("survivors").c_str());

    int couple = 0;
    for(int ind=0; fitest[ind] != -1; ind++){
        Genome *female = new Genome();
        female->Load(fitest[ind]);
        Genome *male = new Genome();
        ind++;
        if(fitest[ind] == -1) break;
        male->Load(fitest[ind]);
        for(int i=0; i<childsPerPair; i++){
            Genome *child = female->Mate(male);
            LOGGER(2, "Mated " << female->GetId() << " with " << male->GetId() << " ===> " << child->GetId());
            child->Save();
            delete child;
        }
        if(saveBest > couple){
            female->Save();
            male->Save();
        } else {
            // here, parents die after mating.
            female->Die();
            male->Die();
        }
        delete female;
        delete male;
        couple += 2;
    }
    delete [] fitest;
    GFile::EmptyDir(Genome::GetAttribute("geneserver_dir") + "/incoming/");
}


vector <Genome::ChromosomeFormat> Genome::StringToFormat(string fmt){
    unsigned int lbracket = 0;
    unsigned int rbracket, comma1, comma2;

    vector <Genome::ChromosomeFormat> chromosomeFormat;
    while ((lbracket = fmt.find('{', lbracket)) != string::npos){
        lbracket++;
        comma1 = fmt.find(',', lbracket);
        comma2 = fmt.find(',', comma1+1);
        if(comma1 == string::npos || comma2 == string::npos) throw runtime_error("error in format string\n");
        rbracket = fmt.find('}', lbracket);
        if(rbracket == string::npos) throw runtime_error("error in format string\n");
        Genome::ChromosomeFormat chr;
        chr.genes = atoi(fmt.substr(lbracket, comma1).c_str());
        chr.gene_size = atoi(fmt.substr(comma1+1, rbracket).c_str());
        chr.mutationMode = (unsigned char) fmt[comma2+1];
        chromosomeFormat.push_back(chr);
    }
    return  chromosomeFormat;
}
